#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
from prjxray import verilog
import json

# Set to true to enable additional tags useful for tracing bit toggles.
DEBUG_FUZZER = False


def bitfilter(frame, word):
    if frame < 30 or frame > 37:
        return False

    return True


def handle_data_width(segmk, d):
    if 'DATA_WIDTH' not in d:
        return

    site = d['ologic_loc']

    data_rate = verilog.unquote(d['DATA_RATE_OQ'])
    segmk.add_site_tag(
        site, 'OSERDES.DATA_WIDTH.{}.W{}'.format(data_rate, d['DATA_WIDTH']),
        1)


def no_oserdes(segmk, site):
    for mode in ['SDR', 'DDR']:
        if mode == 'SDR':
            widths = [2, 3, 4, 5, 6, 7, 8]
        else:
            assert mode == 'DDR'
            widths = [4, 6, 8, 10, 14]

        for opt in widths:
            segmk.add_site_tag(
                site, 'OSERDES.DATA_WIDTH.{}.W{}'.format(mode, opt), 0)


def main():
    print("Loading tags")
    segmk = Segmaker("design.bits")

    with open('params.jl', 'r') as f:
        design = json.load(f)

        for d in design:
            site = d['ologic_loc']

            handle_data_width(segmk, d)

            segmk.add_site_tag(site, 'OSERDES.IN_USE', d['use_oserdese2'])

            if d['use_oserdese2']:
                segmk.add_site_tag(site, 'OQUSED', 1)

                for opt in ['SDR', 'DDR']:
                    segmk.add_site_tag(
                        site, 'OSERDES.DATA_RATE_OQ.{}'.format(opt),
                        verilog.unquote(d['DATA_RATE_OQ']) == opt)

                data_rate_tq = verilog.unquote(d['DATA_RATE_TQ'])
                segmk.add_site_tag(
                    site, 'OSERDES.DATA_RATE_TQ.{}'.format(data_rate_tq), 1)
                for opt in ['BUF', 'SDR', 'DDR']:
                    segmk.add_site_tag(
                        site, 'OSERDES.DATA_RATE_TQ.{}'.format(opt),
                        opt == data_rate_tq)

                for opt in ['SRVAL_OQ', 'SRVAL_TQ', 'INIT_OQ', 'INIT_TQ']:
                    segmk.add_site_tag(site, opt, d[opt])
                    segmk.add_site_tag(site, 'Z' + opt, 1 ^ d[opt])

                for opt in ['CLK', 'CLKDIV']:
                    if d['{}_USED'.format(opt)]:
                        k = 'IS_{}_INVERTED'.format(opt)
                        segmk.add_site_tag(site, k, d[k])
                        segmk.add_site_tag(
                            site, 'ZINV_{}'.format(opt), 1 ^ d[k])

                if d['io']:
                    for idx in range(4):
                        k = 'IS_T{}_INVERTED'.format(idx + 1)
                        segmk.add_site_tag(site, k, d[k])
                        segmk.add_site_tag(
                            site, 'ZINV_T{}'.format(idx + 1), 1 ^ d[k])

                for idx in range(8):
                    k = 'IS_D{}_INVERTED'.format(idx + 1)
                    segmk.add_site_tag(site, k, d[k])
                    segmk.add_site_tag(
                        site, 'ZINV_D{}'.format(idx + 1), 1 ^ d[k])

                for tristate_width in [1, 4]:
                    segmk.add_site_tag(
                        site,
                        'OSERDES.TRISTATE_WIDTH.W{}'.format(tristate_width),
                        d['TRISTATE_WIDTH'] == tristate_width)

                for opt in ['MASTER', 'SLAVE']:
                    segmk.add_site_tag(
                        site, 'OSERDES.SERDES_MODE.{}'.format(opt),
                        opt == verilog.unquote(d['OSERDES_MODE']))

            if 'o_sr_used' in d:
                if d['o_sr_used'] in ['S', 'R']:
                    segmk.add_site_tag(site, 'ODDR.SRUSED', 1)
                    segmk.add_site_tag(site, 'ODDR.ZSRUSED', 0)
                else:
                    assert d['o_sr_used'] == 'None'
                    segmk.add_site_tag(site, 'ODDR.SRUSED', 0)
                    segmk.add_site_tag(site, 'ODDR.ZSRUSED', 1)

            if 't_sr_used' in d:
                if d['t_sr_used'] in ['S', 'R']:
                    segmk.add_site_tag(site, 'TDDR.SRUSED', 1)
                    segmk.add_site_tag(site, 'TDDR.ZSRUSED', 0)
                else:
                    assert d['t_sr_used'] == 'None'
                    segmk.add_site_tag(site, 'TDDR.SRUSED', 0)
                    segmk.add_site_tag(site, 'TDDR.ZSRUSED', 1)

            if d['oddr_mux_config'] == 'direct':
                segmk.add_site_tag(site, 'ODDR_TDDR.IN_USE', 1)

            if d['tddr_mux_config'] == 'direct':
                segmk.add_site_tag(site, 'ODDR_TDDR.IN_USE', 1)

            if d['oddr_mux_config'] == 'direct' and d[
                    'tddr_mux_config'] == 'direct':
                segmk.add_site_tag(site, 'ZINV_CLK', 1 ^ d['IS_CLK_INVERTED'])

                if d['IS_CLK_INVERTED'] == 0:
                    for opt in ['OPPOSITE_EDGE', 'SAME_EDGE']:
                        segmk.add_site_tag(
                            site, 'ODDR.DDR_CLK_EDGE.{}'.format(opt),
                            verilog.unquote(d['ODDR_CLK_EDGE']) == opt)

                segmk.add_site_tag(
                    site, 'TDDR.DDR_CLK_EDGE.INV',
                    d['ODDR_CLK_EDGE'] != d['TDDR_CLK_EDGE'])
                segmk.add_site_tag(
                    site, 'TDDR.DDR_CLK_EDGE.ZINV',
                    d['ODDR_CLK_EDGE'] == d['TDDR_CLK_EDGE'])

                if 'SRTYPE' in d:
                    for opt in ['ASYNC', 'SYNC']:
                        segmk.add_site_tag(
                            site, 'OSERDES.SRTYPE.{}'.format(opt),
                            verilog.unquote(d['SRTYPE']) == opt)

                for opt in ['ASYNC', 'SYNC']:
                    segmk.add_site_tag(
                        site, 'OSERDES.TSRTYPE.{}'.format(opt),
                        verilog.unquote(d['TSRTYPE']) == opt)

            if not d['use_oserdese2']:
                no_oserdes(segmk, site)
                if d['oddr_mux_config'] == 'lut':
                    segmk.add_site_tag(site, 'ODDR_TDDR.IN_USE', 0)
                    segmk.add_site_tag(site, 'OMUX.D1', 1)
                    segmk.add_site_tag(site, 'OQUSED', 1)
                elif d['oddr_mux_config'] == 'direct':
                    segmk.add_site_tag(site, 'OMUX.D1', 0)
                elif d['oddr_mux_config'] == 'none' and not d['io']:
                    segmk.add_site_tag(site, 'OQUSED', 0)

            segmk.add_site_tag(site, 'TQUSED', d['io'])

            if DEBUG_FUZZER:
                for k in d:
                    segmk.add_site_tag(
                        site, 'param_' + k + '_' + str(d[k]).replace(
                            ' ', '').replace('\n', ''), 1)

    segmk.compile(bitfilter=bitfilter)
    segmk.write(allow_empty=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import os
import random
from collections import namedtuple

random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.lut_maker import LutMaker
from prjxray.db import Database

INT = "INT"
BIN = "BIN"


def gen_sites(tile, site, filter_cmt=None):
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        if tile not in gridinfo.tile_type:
            continue
        else:
            tile_type = gridinfo.tile_type

        for site_name, site_type in gridinfo.sites.items():
            if site_type != site:
                continue

            cmt = gridinfo.clock_region

            if filter_cmt is not None and cmt != filter_cmt:
                continue

            yield tile_name, tile_type, site_name, cmt


def main():
    print(
        '''
module top(
    input wire in,
    output wire out
);

assign out = in;
''')

    luts = LutMaker()

    params_dict = {"tile_type": None}
    params_list = list()

    clkswing_cfg_tiles = dict()
    ibufds_out_wires = dict()
    for tile_name, _, site_name, _ in gen_sites("GTP_COMMON", "IBUFDS_GTE2"):
        # Both the IBUFDS_GTE2 in the same tile need to have
        # the same CLKSWING_CFG parameter
        if tile_name not in clkswing_cfg_tiles:
            clkswing_cfg = random.randint(0, 3)
            clkswing_cfg_tiles[tile_name] = clkswing_cfg
        else:
            clkswing_cfg = clkswing_cfg_tiles[tile_name]

        in_use = bool(random.randint(0, 9))
        params = {
            "site":
            site_name,
            "tile":
            tile_name,
            "IN_USE":
            in_use,
            "CLKRCV_TRST":
            verilog.quote("TRUE" if random.randint(0, 1) else "FALSE"),
            "CLKCM_CFG":
            verilog.quote("TRUE" if random.randint(0, 1) else "FALSE"),
            "CLKSWING_CFG":
            clkswing_cfg,
        }

        if in_use:
            ibufds_out_wire = "{}_O".format(site_name)

            if tile_name not in ibufds_out_wires:
                ibufds_out_wires[tile_name] = list()

            ibufds_out_wires[tile_name].append(
                (ibufds_out_wire, int(site_name[-1]) % 2))

            print("wire {};".format(ibufds_out_wire))
            print("(* KEEP, DONT_TOUCH, LOC=\"{}\" *)".format(site_name))
            print(
                """
IBUFDS_GTE2 #(
    .CLKRCV_TRST({CLKRCV_TRST}),
    .CLKCM_CFG({CLKCM_CFG}),
    .CLKSWING_CFG({CLKSWING_CFG})
) {site} (
    .O({out})
);""".format(**params, out=ibufds_out_wire))

        params_list.append(params)

    DRP_PORTS = [
        ("DRPCLK", "clk"), ("DRPEN", "in"), ("DRPWE", "in"), ("DRPRDY", "out")
    ]

    for tile_name, tile_type, site_name, cmt in gen_sites("GTP_COMMON",
                                                          "GTPE2_COMMON"):

        params_dict["tile_type"] = tile_type

        params = dict()
        params['site'] = site_name
        params['tile'] = tile_name

        verilog_attr = ""

        verilog_attr = "#("

        fuz_dir = os.getenv("FUZDIR", None)
        assert fuz_dir
        with open(os.path.join(fuz_dir, "attrs.json"), "r") as attrs_file:
            attrs = json.load(attrs_file)

        in_use = bool(random.randint(0, 9))
        params["IN_USE"] = in_use

        if in_use:
            for param, param_info in attrs.items():
                param_type = param_info["type"]
                param_values = param_info["values"]
                param_digits = param_info["digits"]

                if param_type == INT:
                    value = random.choice(param_values)
                    value_str = value
                else:
                    assert param_type == BIN
                    value = random.randint(0, param_values[0])
                    value_str = "{digits}'b{value:0{digits}b}".format(
                        value=value, digits=param_digits)

                params[param] = value

                verilog_attr += """
            .{}({}),""".format(param, value_str)

            verilog_ports = ""

            for param in ["PLL0LOCKDETCLK", "PLL1LOCKDETCLK", "DRPCLK"]:
                is_inverted = random.randint(0, 1)

                params[param] = is_inverted

                verilog_attr += """
            .IS_{}_INVERTED({}),""".format(param, is_inverted)
                verilog_ports += """
            .{}({}),""".format(param, luts.get_next_output_net())

            verilog_attr = verilog_attr.rstrip(",")
            verilog_attr += "\n)"

            for param in ["GTREFCLK0_USED", "GTREFCLK1_USED",
                          "BOTH_GTREFCLK_USED"]:
                params[param] = 0

            if tile_name in ibufds_out_wires:
                gtrefclk_ports_used = 0

                for wire, location in ibufds_out_wires[tile_name]:
                    if random.random() < 0.5:
                        continue

                    verilog_ports += """
            .GTREFCLK{}({}),""".format(location, wire)

                    gtrefclk_ports_used += 1
                    params["GTREFCLK{}_USED".format(location)] = 1

                if gtrefclk_ports_used == 2:
                    params["BOTH_GTREFCLK_USED"] = 1

            enable_drp = random.randint(0, 1)
            params["ENABLE_DRP"] = enable_drp

            for _, _, channel_site_name, _ in gen_sites("GTP_CHANNEL",
                                                        "GTPE2_CHANNEL", cmt):

                if not enable_drp:
                    break

                verilog_ports_channel = ""
                for port, direction in DRP_PORTS:
                    if direction == "in":
                        verilog_ports_channel += """
    .{}({}),""".format(port, luts.get_next_output_net())

                    elif direction == "clk":
                        # DRPCLK needs to come from a clock source
                        print(
                            """
wire clk_bufg_{site};

(* KEEP, DONT_TOUCH *)
BUFG bufg_{site} (.O(clk_bufg_{site}));""".format(site=channel_site_name))

                        verilog_ports_channel += """
    .{}(clk_bufg_{}),""".format(port, channel_site_name)

                    elif direction == "out":
                        verilog_ports_channel += """
    .{}({}),""".format(port, luts.get_next_input_net())

                print(
                    """
(* KEEP, DONT_TOUCH, LOC=\"{site}\" *)
GTPE2_CHANNEL {site} (
    {ports}
);""".format(ports=verilog_ports_channel.rstrip(","), site=channel_site_name))

            print(
                """
(* KEEP, DONT_TOUCH, LOC=\"{site}\" *)
GTPE2_COMMON {attrs} {site} (
    {ports}
);""".format(
                    attrs=verilog_attr,
                    ports=verilog_ports.rstrip(","),
                    site=site_name))

        params_list.append(params)

    for l in luts.create_wires_and_luts():
        print(l)

    print("endmodule")

    params_dict["params"] = params_list
    with open('params.json', 'w') as f:
        json.dump(params_dict, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This script generates code snippets for cell definition, technology mapping and
VPR architecture definition. It reads correlation of IOB fasm features with
IO standard settings such as IOSTANDARD, DRIVE and SLEW and generates:

- Verilog model parameter definitions that match fasm features.
- Xilinx's IOB parameter translation to those features.
- Assignment of model parameters to the actual fasm features for VPR arch def.

Generated code snippets need to be manually copied to relevant places in the
target code / arch. def.
"""
import os
import argparse
import csv

from collections import namedtuple

IOSettings = namedtuple("IOSettings", "iostandard, drive, slew, is_diff")

# =============================================================================


def load_feature_data(fname):
    """
    Load feature vs. IO settings correlation data from a CSV file.
    """

    # Load the data from CSV
    feature_data = {}
    with open(fname, "r") as fp:
        for line in csv.DictReader(fp):

            # Get IOSettings
            ios_parts = line["iosettings"].split(".")
            ios = IOSettings(
                ios_parts[0],
                int(ios_parts[1][1:]) if ios_parts[1][1:] != "_FIXED" else
                None, ios_parts[2], "DIFF_" in ios_parts[0])

            # Feature activity
            feature_data[ios] = {k: line[k] for k in list(line.keys())[1:]}

    return feature_data


def filter_feature_data(feature_data):
    """
    Filters fasm features and leaves only those related to IO settings.
    """

    for data in feature_data.values():
        for feature in dict(data):

            if "DISABLE" in feature:
                del data[feature]
            if "PULLTYPE" in feature:
                del data[feature]
            if "IN_TERM" in feature:
                del data[feature]
            if "LOW_PWR" in feature:
                del data[feature]

    return feature_data


def get_relevant_parameters(feature_data, iob_type):
    """
    Gets features relevant to a specific IOB type and assign them verilog
    parameter names.
    """

    # Get all features and make parameter names
    all_features = set(
        [f for data in feature_data.values() for f in data.keys()])
    parameters = []

    # Check if a feature is relevant to particular IOB type
    is_diff = "DIFF_" in iob_type

    for feature in all_features:
        for iosettings in feature_data:
            if iosettings.is_diff == is_diff:
                base_iob_type = iob_type.replace("DIFF_", "")
                if base_iob_type in feature_data[iosettings][feature]:
                    parameters.append((
                        feature,
                        feature.replace(".", "_"),
                    ))
                    break

    return parameters


# =============================================================================


def generate_parameter_defs(parameters):
    """
    Generates verilog parameter definitions for the IOB cell model.
    """
    verilog = []

    for feature, parameter in sorted(parameters):
        verilog.append("parameter [0:0] {} = 1'b0;".format(parameter))

    return "\n".join(verilog)


def generate_parameter_map(parameters):
    """
    Generates mapping of parameters for fasm features for architecture
    definition.
    """
    xml = []

    xml.append("<meta name=\"fasm_params\">")
    for feature, parameter in sorted(parameters):
        xml.append("  {} = {}".format(feature, parameter))
    xml.append("</meta>")

    return "\n".join(xml)


def generate_parameter_assignments(parameters, feature_data, iob_type):
    """
    Generates parameter assignmets to be used inside the techmap. For
    each IOB cell parameter a set of conditions is generated based on
    feature activity observation.
    """
    verilog = []

    is_diff = "DIFF_" in iob_type
    base_iob_type = iob_type.replace("DIFF_", "")

    for feature, parameter in sorted(parameters):
        condition = []
        for iosettings in feature_data:

            if iosettings.is_diff != is_diff:
                continue

            # Feature is set
            if base_iob_type in feature_data[iosettings][feature]:
                cond = "IOSTANDARD == \"{}\"".format(
                    iosettings.iostandard.upper())

                if base_iob_type in ["O", "T"]:
                    if iosettings.drive is not None and "DRIVE" in feature:
                        cond += " && DRIVE == {}".format(iosettings.drive)
                    if "SLEW" in feature:
                        cond += " && SLEW == \"{}\"".format(iosettings.slew)

                condition.append(cond)

        condition = sorted(list(set(condition)))

        condition_str = " || \n".join(["  ({})".format(c) for c in condition])
        verilog.append(".{}(\n{}\n)".format(parameter, condition_str))

    return ",\n".join(verilog)


# =============================================================================


def run():
    """
    Main.
    """

    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("features", type=str, help="Input CSV file")

    args = parser.parse_args()

    # Load iosettings vs. feature map
    feature_data = load_feature_data(args.features)
    # Reject non-iostandard related features
    feature_data = filter_feature_data(feature_data)

    # Make parameters
    parameters = {}
    for iob_type in ["I", "O", "T", "DIFF_I", "DIFF_O", "DIFF_T"]:
        parameters[iob_type] = get_relevant_parameters(feature_data, iob_type)

    # Parameter definition
    with open("cells_sim.v", "w") as fp:
        for iob_type, params in parameters.items():
            fp.write("// {} {}\n\n".format(iob_type, "=" * 70))
            verilog = generate_parameter_defs(params)
            fp.write(verilog + "\n\n")

    # Parameter assignments
    with open("cells_map.v", "w") as fp:
        for iob_type, params in parameters.items():
            fp.write("// {} {}\n\n".format(iob_type, "=" * 70))
            verilog = generate_parameter_assignments(
                params, feature_data, iob_type)
            fp.write(verilog + "\n\n")

    # Parameter assignments
    with open("arch.xml", "w") as fp:
        for iob_type, params in parameters.items():
            fp.write("<!-- {} -->\n\n".format(iob_type))
            xml = generate_parameter_map(params)
            fp.write(xml + "\n\n")


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This is a db fixup script that does two things:

1. Clears (removes) all bits found in "IN_USE" tag(s) and removes the IN_USE
   tag itself

2. Reads tag group definition from a file and applies the tag grouping. First
   a set of common bits for each group is found (logical OR among all tags
   belonging to the group). Then in each tag belonging to the group those
   bits are set to 0 but only if they are not already present there.

The resulting data is written into a segbits file.
"""
import argparse
import re
import itertools

# =============================================================================


def load_tag_groups(file_name):
    """
    Loads tag groups from a text file.

    A tag group is defined by specifying a space separated list of tags within
    a single line. Lines that are empty or start with '#' are ignored.
    """
    tag_groups = []

    # Load tag group specifications
    with open(file_name, "r") as fp:
        for line in fp:
            line = line.strip()

            if len(line) == 0 or line.startswith("#"):
                continue

            group = set(line.split())
            if len(group):
                tag_groups.append(group)

    # Check if all tag groups are exclusive
    for tag_group_a, tag_group_b in itertools.combinations(tag_groups, 2):

        tags = tag_group_a & tag_group_b
        if len(tags):
            raise RuntimeError(
                "Tag(s) {} are present in multiple groups".format(
                    " ".join(tags)))

    return tag_groups


# =============================================================================


def parse_bit(bit):
    """
    Decodes string describing a bit. Returns a tuple (frame, bit, value)
    """
    match = re.match("^(!?)([0-9]+)_([0-9]+)$", bit)
    assert match != None, bit

    val = int(match.group(1) != "!")
    frm = int(match.group(2))
    bit = int(match.group(3))

    return frm, bit, val


def bit_to_str(bit):
    """
    Converts a tuple (frame, bit, value) to its string representation.
    """
    s = "!" if not bit[2] else ""
    return "{}{}_{:02d}".format(s, bit[0], bit[1])


def load_segbits(file_name):
    """
    Loads a segbits file.
    """

    segbits = {}

    with open(file_name, "r") as fp:
        for line in fp:
            line = line.strip()
            fields = line.split()

            if len(fields) < 2:
                raise RuntimeError("Malformed line: '%s'" % line)

            tag = fields[0]

            if "<" in line or ">" in line:
                segbits[tag] = set()

            else:
                bits = set([parse_bit(bit) for bit in fields[1:]])
                segbits[tag] = bits

    return segbits


def save_segbits(file_name, segbits):
    """
    Save segbits to a .db or .rdb file
    """

    with open(file_name, "w") as fp:
        for tag, bits in segbits.items():
            if not len(bits):
                continue

            line = tag + " "
            line += " ".join([bit_to_str(bit) for bit in sorted(list(bits))])
            fp.write(line + "\n")


# =============================================================================


def mask_out_bits(segbits, mask, tags_to_mask=None):
    """
    Given a set of bits and a list of tags to affect (optional) removes all
    the bits from each tag that are present (and equal) in the masking set.
    """

    if tags_to_mask is None:
        tags_to_mask = segbits.keys()

    # Mask out matching bits
    for tag in tags_to_mask:
        bits = segbits[tag]

        bits = set(bits) - set(mask)
        segbits[tag] = bits

    return segbits


def find_common_bits_for_tag_groups(segbits, tag_groups):
    """
    For each tag group finds a common set of bits that have value of one.
    """

    bit_groups = []

    for tag_group in tag_groups:
        bit_group = set()

        for tag, bits in segbits.items():
            if tag in tag_group:
                ones = set([b for b in bits if b[2]])
                bit_group |= ones

        bit_groups.append(bit_group)

    return bit_groups


def group_tags(segbits, tag_groups, bit_groups):
    """
    Implements tag grouping. If a tag belongs to a group then the common bits
    of that group are added to is as zeros.
    """

    for tag_group, bit_group in zip(tag_groups, bit_groups):
        zeros = set([(b[0], b[1], 0) for b in bit_group])
        for tag in tag_group:

            # Insert zero bits
            if tag in segbits.keys():
                bits = segbits[tag]
                for z in zeros:
                    if not (z[0], z[1], 1) in bits:
                        bits.add(z)

    return segbits


# =============================================================================


def main():

    # Parse arguments
    parser = argparse.ArgumentParser()

    parser.add_argument("-i", required=True, type=str, help="Input .rdb file")
    parser.add_argument(
        "-g", required=True, type=str, help="Input tag group definition file")
    parser.add_argument("-o", required=True, type=str, help="Output .rdb file")

    args = parser.parse_args()

    # Load tag groups
    tag_groups = load_tag_groups(args.g)

    # Load raw database file
    segbits = load_segbits(args.i)

    # Mask out bits present in "IN_USE" tag(s) as they are common to other tags
    for tag in segbits.keys():
        if tag.endswith("IN_USE"):
            prefix = tag[:tag.index("IN_USE")]
            tags_to_mask = [t for t in segbits.keys() if t.startswith(prefix)]
            mask_out_bits(segbits, segbits[tag], tags_to_mask)

    tags_to_remove = set()
    for tag, bits in segbits.items():
        if len(bits) == 0:
            tags_to_remove.add(tag)

    for tag in tags_to_remove:
        del segbits[tag]

    for tag in segbits.keys():
        if tag.endswith("_ACTIVE") and 'FREQ_BB' in tag:
            m = re.search('FREQ_BB([0-9])', tag)
            l_prefix = '.CMT_TOP_L_UPPER_T_FREQ_BB{}'.format(m.group(1))
            r_prefix = '.CMT_TOP_R_UPPER_T_FREQ_BB{}'.format(m.group(1))

            l_tags_to_mask = [t for t in segbits.keys() if t.endswith(l_prefix)]
            r_tags_to_mask = [t for t in segbits.keys() if t.endswith(r_prefix)]

            mask_out_bits(segbits, segbits[tag], l_tags_to_mask)
            mask_out_bits(segbits, segbits[tag], r_tags_to_mask)

    # Find common bits
    bit_groups = find_common_bits_for_tag_groups(segbits, tag_groups)
    # Apply tag grouping
    segbits = group_tags(segbits, tag_groups, bit_groups)

    # Save fixed database file
    save_segbits(args.o, segbits)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker
from prjxray import verilog


def main():
    segmk = Segmaker("design.bits")

    print("Loading tags")
    with open('params.json') as f:
        params = json.load(f)

    for row in params:
        base_name = 'BUFHCE_X{}Y{}'.format(row['x'], row['y'])

        segmk.add_site_tag(
            row['site'], '{}.IN_USE'.format(base_name), row['IN_USE'])
        if not row['IN_USE']:
            continue

        segmk.add_site_tag(
            row['site'], '{}.INIT_OUT'.format(base_name), row['INIT_OUT'])

        segmk.add_site_tag(
            row['site'], '{}.ZINV_CE'.format(base_name),
            1 ^ row['IS_CE_INVERTED'])

        # SYNC is a zero pattern
        for opt in ['ASYNC']:
            segmk.add_site_tag(
                row['site'], '{}.CE_TYPE.'.format(base_name) + opt,
                verilog.unquote(row['CE_TYPE']) == opt)

    segmk.compile()
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import re

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

tiledata = dict()
pipdata = dict()
ignpip = set()

print("Loading tags from design.txt.")
with open("design.txt", "r") as f:
    for line in f:
        tile, pip, src, dst, pnum, pdir = line.split()
        _, pip = pip.split(".")
        _, src = src.split("/")
        _, dst = dst.split("/")
        pnum = int(pnum)
        pdir = int(pdir)

        if tile not in tiledata:
            tiledata[tile] = {"pips": set(), "srcs": set(), "dsts": set()}

        if pip in pipdata:
            assert pipdata[pip] == (src, dst)
        else:
            pipdata[pip] = (src, dst)

        tiledata[tile]["pips"].add(pip)
        tiledata[tile]["srcs"].add(src)
        tiledata[tile]["dsts"].add(dst)

        if pdir == 0:
            tiledata[tile]["srcs"].add(dst)
            tiledata[tile]["dsts"].add(src)

        if pnum == 1 or pdir == 0 or not re.match(r"^(VCC|GND)_WIRE$", src):
            ignpip.add(pip)

for tile, pips_srcs_dsts in tiledata.items():
    pips = pips_srcs_dsts["pips"]
    srcs = pips_srcs_dsts["srcs"]
    dsts = pips_srcs_dsts["dsts"]

    for pip, src_dst in pipdata.items():
        src, dst = src_dst
        if pip in ignpip:
            pass
        elif pip in pips:
            segmk.addtag(tile, "%s.%s" % (dst, src), 1)
        elif src_dst[1] not in dsts:
            segmk.addtag(tile, "%s.%s" % (dst, src), 0)

segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from __future__ import print_function
import sys, re
import os
import glob
import hashlib


def bytehex(x):
    return ''.join('{:02x}'.format(x) for x in x)


def wc_for_iteration(todo_dir, fni):
    with open("%s/%u_all.txt" % (todo_dir, fni), "rb") as f:
        return sum(1 for _ in f)


def check_made_progress(todo_dir, max_iter, min_progress):
    """ Returns true if minimum progress is being made. """
    if max_iter == 1:
        return True

    prev_iteration = wc_for_iteration(todo_dir, max_iter - 1)
    cur_iteration = wc_for_iteration(todo_dir, max_iter)

    made_progress = prev_iteration - cur_iteration > min_progress
    if not made_progress:
        print(
            "Between iteration {} and iteration {} only {} pips were solved.  Terminating iteration."
            .format(max_iter - 1, max_iter, prev_iteration - cur_iteration))

    return made_progress


def run(
        todo_dir,
        min_iters=None,
        min_progress=None,
        timeout_iters=None,
        max_iters=None,
        zero_entries=None,
        zero_entries_filter=".*",
        verbose=False):
    timeout_fn = "%s/timeout" % todo_dir
    # make clean removes todo dir, but helps debugging
    if os.path.exists(timeout_fn):
        print("WARNING: removing %s" % timeout_fn)
        os.remove(timeout_fn)

    alls = glob.glob("%s/*_all.txt" % todo_dir)
    max_iter = 0
    for fn in alls:
        n = int(re.match(r".*/([0-9]*)_all.txt", fn).group(1))
        max_iter = max(max_iter, n)

    if max_iter == 0:
        print("Incomplete: no iters")
        sys.exit(1)

    verbose and print("Max iter: %u, need: %s" % (max_iter, min_iters))

    # Don't allow early termination if below min_iters
    if min_iters is not None and max_iter < min_iters:
        print("Incomplete: not enough iters")
        sys.exit(1)

    # Force early termination if at or above max_iters.
    if max_iters is not None and max_iter >= max_iters:
        print(
            "Complete: reached max iters (want %u, got %u)" %
            (max_iters, max_iter))
        sys.exit(0)

    # Mark timeout if above timeout_iters
    if timeout_iters is not None and max_iter > timeout_iters:
        print("ERROR: timeout (max %u, got %u)" % (timeout_iters, max_iter))
        with open(timeout_fn, "w") as _f:
            pass
        sys.exit(1)

    # Check if zero entries criteria is not met.
    if zero_entries:
        filt = re.compile(zero_entries_filter)
        count = 0
        fn = "%s/%u_all.txt" % (todo_dir, max_iter)
        with open(fn, 'r') as f:
            for l in f:
                if filt.search(l):
                    count += 1

        if count > 0:
            print("%s: %s lines" % (fn, count))
            print(
                "Incomplete: need zero entries (used filter: {})".format(
                    repr(zero_entries_filter)))
            sys.exit(1)
        else:
            # If there are zero entries, check if min_progress criteria is in
            # affect. If so, that becomes the new termination condition.
            if min_progress is None:
                print(
                    "No unfiltered entries, done (used filter: {})!".format(
                        repr(zero_entries_filter)))
                sys.exit(0)
            else:
                # Even if there are 0 unfiltered entries, fuzzer may still be
                # making progress with filtered entries.
                print(
                    "No unfiltered entries (used filter: {}), checking if progress is being made"
                    .format(repr(zero_entries_filter)))

    # Check if minimum progress was achieved, continue iteration if so.
    if min_progress is not None and not check_made_progress(todo_dir, max_iter,
                                                            min_progress):
        sys.exit(0)

    print("No exit criteria met, keep going!")
    sys.exit(1)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description=
        "Check int_loop completion. Exits 0 on done, 1 if more loops are needed"
    )

    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument('--todo-dir', default="build/todo", help='')
    parser.add_argument(
        '--min-iters', default=None, help='Minimum total number of iterations')
    parser.add_argument(
        '--min-progress',
        default=None,
        help=
        'Minimum amount of process between iterations.  If less progress is made, terminates immediately.'
    )
    parser.add_argument(
        '--timeout-iters',
        default=None,
        help='Max number of entries before creating todo/timeout')
    parser.add_argument(
        '--max-iters',
        default=None,
        help='Max number of entries before declaring success')
    parser.add_argument(
        '--zero-entries',
        action="store_true",
        help='Must be no unsolved entries in latest')
    parser.add_argument(
        '--zero-entries-filter',
        default=".*",
        help=
        'When zero-entries is supplied, this filter is used to filter pips used for counting against zero entries termination condition.'
    )
    args = parser.parse_args()

    def zint(x):
        return None if x is None else int(x)

    run(
        todo_dir=args.todo_dir,
        min_iters=zint(args.min_iters),
        min_progress=zint(args.min_progress),
        timeout_iters=zint(args.timeout_iters),
        max_iters=zint(args.max_iters),
        zero_entries=args.zero_entries,
        zero_entries_filter=args.zero_entries_filter,
        verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
NOTE: "segments" as used in this file is mostly unrelated to tilegrid.json usage
ie tilegrid.json has names like SEG_CLBLL_L_X2Y50 where as here they are tile based and named like seg_00400100_02
Instead of using tilegrid.json "segments, segments are formed by looking for tiles that use the same address + offset

Sample segdata.txt output (from 015-clbnffmux/specimen_001/segdata_clbll_r.txt):
seg 00020880_048
bit 30_00
bit 31_49
tag CLB.SLICE_X0.AFF.DMUX.CY 1
tag CLB.SLICE_X0.BFF.DMUX.BX 0

tilegrid.json provides tile addresses
'''

import os, json, re
from prjxray.util import OpenSafeFile, get_db_root, get_fabric

BLOCK_TYPES = set(('CLB_IO_CLK', 'BLOCK_RAM', 'CFG_CLB'))


def recurse_sum(x):
    '''Count number of nested iterable occurances'''
    if type(x) in (str, bytearray):
        return 1
    if type(x) in (dict, ):
        return sum([recurse_sum(y) for y in x.values()])
    else:
        try:
            return sum([recurse_sum(y) for y in x])
        except TypeError:
            return 1


def json_hex2i(s):
    '''Convert a JSON hex literal into an integer (it can't store hex natively)'''
    # TODO: maybe just do int(x, 0)
    return int(s[2:], 16)


def add_site_group_zero(segmk, site, prefix, vals, zero_val, val):
    '''
    Correctly add tags for a multi-bit enumerated value
    Naively adding them directly doesn't work correctly because overlapping bits won't solve correctly
    Instead, you need to carefully diff against a known zero value
    Good zero values:
    -An enum known to be zero
    -A site that doesn't contain the enum

    segmak: Segmaker object
    site: the site to add tags to
    prefix: tag string to prefix onto vals
    vals: all possible tag enum vals
    zero_val: tag value known to have no bits set
    '''
    # assert zero_val in vals, "Got %s, need %s" % (zero_val, vals)
    assert val in vals or val == zero_val, "Got %s, need %s" % (val, vals)

    if val == zero_val:
        # Zero symbol occured, none of the others did
        for aval in vals:
            tag = prefix + aval
            segmk.add_site_tag(site, tag, aval == val)
    else:
        # Only add the occured symbol
        tag = prefix + val
        segmk.add_site_tag(site, tag, True)
        if zero_val in vals:
            # And zero so that it has something to solve against
            tag = prefix + zero_val
            segmk.add_site_tag(site, tag, False)


class Segmaker:
    def __init__(self, bitsfile, verbose=None, db_root=None, fabric=None):
        self.db_root = db_root
        if self.db_root is None:
            self.db_root = get_db_root()
            assert self.db_root, "No db root specified."

        self.fabric = fabric
        if self.fabric is None:
            self.fabric = get_fabric()
            assert self.fabric, "No fabric specified."

        self.verbose = verbose if verbose is not None else os.getenv(
            'VERBOSE', 'N') == 'Y'
        self.load_grid()
        self.load_bits(bitsfile)
        '''
        self.tags[site][name] = value
        Where:
        -site: ex 'SLICE_X13Y101'
        -name: ex 'CLB.SLICE_X0.AFF.DMUX.CY'
        '''
        self.site_tags = dict()
        self.tile_tags = dict()

        # output after compiling
        self.segments_by_type = None

        # hacky...improve if we encounter this more
        self.def_bt = 'CLB_IO_CLK'
        self.index_sites()

    def index_sites(self):
        self.verbose and print("Indexing sites")
        self.sites = {}
        for tilename, tiledata in self.grid.items():
            for site in tiledata["sites"]:
                self.sites[site] = tilename
        self.verbose and print("Sites indexed")

    def set_def_bt(self, block_type):
        '''Set default block type when more than one block present'''
        assert block_type in BLOCK_TYPES, (
            "Unknown block type %r (known %r)" % (block_type, BLOCK_TYPES))
        self.def_bt = block_type

    def load_grid(self):
        '''Load self.grid holding tile addresses'''
        with OpenSafeFile(os.path.join(self.db_root, self.fabric, "tilegrid.json"),
                  "r") as f:
            self.grid = json.load(f)
        assert "segments" not in self.grid, "Old format tilegrid.json"

    def load_bits(self, bitsfile):
        '''Load self.bits holding the bits that occured in the bitstream'''
        '''
        Format:
        self.bits[base_frame][bit_wordidx] = set()
        Where elements are (bit_frame, bit_wordidx, bit_bitidx))
        bit_frame is a relatively large number forming the FDRI address
        base_frame is a truncated bit_frame address of related FDRI addresses
        0 <= bit_wordidx <= 100
        0 <= bit_bitidx < 31

        Sample bits input
        bit_00020500_000_08
        bit_00020500_000_14
        bit_00020500_000_17
        '''
        self.bits = dict()
        print("Loading bits from %s." % bitsfile)
        with OpenSafeFile(bitsfile, "r") as f:
            for line in f:
                # ex: bit_00020500_000_17
                line = line.split("_")
                bit_frame = int(line[1], 16)
                bit_wordidx = int(line[2], 10)
                bit_bitidx = int(line[3], 10)
                base_frame = bit_frame & ~0x7f

                self.bits.setdefault(base_frame, dict()).setdefault(
                    bit_wordidx, set()).add(
                        (bit_frame, bit_wordidx, bit_bitidx))
        if self.verbose:
            print(
                'Loaded bits: %u bits in %u base frames' %
                (recurse_sum(self.bits), len(self.bits)))

    def add_site_tag(self, site, name, value):
        '''
        XXX: can add tags in two ways:
        -By site name
        -By tile name (used for pips?)
        Consider splitting into two separate data structures

        Record, according to value, if (site, name) exists

        Ex:
        self.addtag('SLICE_X13Y101', 'CLB.SLICE_X0.AFF.DMUX.CY', 1)
        Indicates that the SLICE_X13Y101 site has an element called 'CLB.SLICE_X0.AFF.DMUX.CY'
        '''
        if '"' in site:
            raise ValueError("Invalid site: %s" % site)
        self.verbose and print(
            'segmaker add tag: site %s tag %s = %s' % (site, name, value))
        assert site in self.sites, "Unknown site %s" % (site, )
        self.site_tags.setdefault(site, dict())[name] = value

    def add_tile_tag(self, tile, name, value):
        # TODO: test this out
        # assert tile in self.grid
        self.verbose and print(
            'segmaker add tag: tile %s tag %s = %s' % (tile, name, value))
        self.tile_tags.setdefault(tile, dict())[name] = value

    def compile(self, bitfilter=None):
        print("Compiling segment data.")
        tags_used = set()
        sites_used = set()
        tile_types_found = set()

        self.segments_by_type = dict()

        def add_segbits(segments, segname, tiledata, bitfilter=None):
            '''
            Add and populate segments[segname]["bits"]
            Gives all of the bits that could exist for the space we are exploring
            Also add segments[segname]["tags"], but don't fill

            segments is a group related to a specific tile type (ex: CLBLM_L)
            It is composed of bits (possible bits) and tags (observed instances)
            segments[segname]["bits"].add(bitname)
            segments[segname]["tags"][tag] = value

            segname: FDRI address + word offset string
            tiledata: tilegrid info for this tile
            '''
            assert segname not in segments
            segment = segments.setdefault(
                segname,
                {
                    "bits": set(),
                    "tags": dict(),
                    # verify new entries match this
                    "offset": bitj["offset"],
                    "words": bitj["words"],
                    "frames": bitj["frames"],
                })

            base_frame = json_hex2i(bitj["baseaddr"])
            for wordidx in range(bitj["offset"],
                                 bitj["offset"] + bitj["words"]):
                if base_frame not in self.bits:
                    continue
                if wordidx not in self.bits[base_frame]:
                    continue
                for bit_frame, bit_wordidx, bit_bitidx in self.bits[
                        base_frame][wordidx]:
                    bitname_frame = bit_frame - base_frame
                    bitname_bit = 32 * (
                        bit_wordidx - bitj["offset"]) + bit_bitidx

                    # Skip bits above the frame limit.
                    if bitname_frame >= bitj["frames"]:
                        continue

                    # some bits are hard to de-correlate
                    # allow force dropping some bits from search space for practicality
                    if bitfilter is None or bitfilter(bitname_frame,
                                                      bitname_bit):
                        bitname = "%02d_%02d" % (bitname_frame, bitname_bit)
                        segment["bits"].add(bitname)

            return segment

        '''
        XXX: wouldn't it be better to iterate over tags? Easy to drop tags
        For now, add a check that all tags are used
        '''
        for tilename, tiledata in self.grid.items():

            def getseg(segname):
                if not segname in segments:
                    return add_segbits(
                        segments, segname, tiledata, bitfilter=bitfilter)
                else:
                    segment = segments[segname]
                    assert segment["offset"] == bitj["offset"]
                    assert segment["words"] == bitj["words"]
                    assert segment["frames"] == bitj["frames"]
                    return segment

            def add_tilename_tags():
                self.verbose and print("Tile %s: check tags" % tilename)
                segment = getseg(segname)

                for name, value in self.tile_tags[tilename].items():
                    tags_used.add((tilename, name))
                    tag = "%s.%s" % (tile_type_norm, name)
                    segment["tags"][tag] = value

            def add_site_tags():
                site_prefix = "_".join(site.split('_')[0:-1])

                def name_slice():
                    '''
                    Simplify SLICE names like:
                    -SLICE_X12Y102 => SLICE_X0
                    -SLICE_X13Y102 => SLICE_X1
                    '''
                    if re.match(r"SLICE_X[0-9]*[02468]Y", site):
                        return "SLICE_X0"
                    elif re.match(r"SLICE_X[0-9]*[13579]Y", site):
                        return "SLICE_X1"
                    else:
                        assert False, "Invalid name in %s" % site

                def name_bram18():
                    # RAMB18_X0Y41
                    if re.match(r"^RAMB18_X.*Y[0-9]*[02468]$", site):
                        return "RAMB18_Y0"
                    elif re.match(r"^RAMB18_X.*Y[0-9]*[13579]$", site):
                        return "RAMB18_Y1"
                    else:
                        assert False, "Invalid name in %s" % site

                def name_y0y1():
                    # RAMB18_X0Y41
                    m = re.match(r"^(.*)_X.*Y[0-9]*[02468]$", site)
                    if m:
                        return "%s_Y0" % m.group(1)

                    m = re.match(r"^(.*)_X.*Y[0-9]*[13579]$", site)
                    if m:
                        return "%s_Y1" % m.group(1)

                    assert 0, site

                def name_default():
                    # most sites are unique within their tile
                    # TODO: maybe verify against DB?
                    return site_prefix

                sitekey = {
                    'SLICE': name_slice,
                    'RAMB18': name_bram18,
                    'IOB': name_y0y1,
                    'IDELAY': name_y0y1,
                    'ODELAY': name_y0y1,
                    'ILOGIC': name_y0y1,
                    'OLOGIC': name_y0y1,
                    'IBUFDS_GTE2': name_y0y1,
                }.get(site_prefix, name_default)()
                self.verbose and print(
                    'site %s w/ %s prefix => tag %s' %
                    (site, site_prefix, sitekey))

                for name, value in self.site_tags[site].items():
                    self.verbose and print("Site %s: check tags" % site)

                    tags_used.add((site, name))
                    tag = "%s.%s.%s" % (tile_type_norm, sitekey, name)
                    # XXX: does this come from name?
                    tag = tag.replace(".SLICEM.", ".")
                    tag = tag.replace(".SLICEL.", ".")
                    segment = getseg(segname)
                    segment["tags"][tag] = value
                sites_used.add(site)

            tile_type = tiledata["type"]
            tile_types_found.add(tile_type)
            segments = self.segments_by_type.setdefault(tile_type, dict())
            '''
            Simplify names by simplifying like:
            -CLBLM_L => CLB
            -CENTER_INTER_R => CENTER_INTER
            -CLK_HROW_TOP_R => CLK_HROW
            -LIOB33 => IOB33
            -LIOI3 => IOI3
            -RIOB18 => IOB18
            -RIOI => IOI
            '''
            tile_type_norm = re.sub("(_TOP|_BOT|LL|LM)?_[LR]$", "", tile_type)
            tile_type_norm = re.sub(
                "_TOP_[LR]_UPPER", "_UPPER", tile_type_norm)

            if tile_type_norm in ['LIOB33', 'RIOB33']:
                tile_type_norm = 'IOB33'
            if tile_type_norm in ['LIOB18', 'RIOB18']:
                tile_type_norm = 'IOB18'
            if tile_type_norm in ['LIOI3', 'RIOI3']:
                tile_type_norm = 'IOI3'
            if tile_type_norm in ['LIOI', 'RIOI']:
                tile_type_norm = 'IOI'
            if tile_type_norm in ['LIOI3_TBYTESRC', 'RIOI3_TBYTESRC']:
                tile_type_norm = 'IOI3'
            if tile_type_norm in ['LIOI3_TBYTETERM', 'RIOI3_TBYTETERM']:
                tile_type_norm = 'IOI3'
            if tile_type_norm in ['LIOI_TBYTESRC', 'RIOI_TBYTESRC']:
                tile_type_norm = 'IOI'
            if tile_type_norm in ['LIOI_TBYTETERM', 'RIOI_TBYTETERM']:
                tile_type_norm = 'IOI'
            if tile_type_norm in ['CMT_TOP_L_LOWER_B', 'CMT_TOP_R_LOWER_B']:
                tile_type_norm = 'CMT_LOWER_B'
            if 'GTP_CHANNEL' in tile_type_norm:
                tile_type_norm = 'GTP_CHANNEL'
            if 'GTP_COMMON' in tile_type_norm:
                tile_type_norm = 'GTP_COMMON'
            if 'GTP_INT_INTERFACE' in tile_type_norm:
                tile_type_norm = 'GTP_INT_INTERFACE'
            if 'GTX_CHANNEL' in tile_type_norm:
                tile_type_norm = 'GTX_CHANNEL'
            if 'GTX_COMMON' in tile_type_norm:
                tile_type_norm = 'GTX_COMMON'
            if 'GTX_INT_INTERFACE' in tile_type_norm:
                tile_type_norm = 'GTX_INT_INTERFACE'

            if tile_type_norm in ['LIOI', 'RIOI']:
                tile_type_norm = 'IOI'
            if tile_type_norm in ['LIOI_TBYTESRC', 'RIOI_TBYTESRC']:
                tile_type_norm = 'IOI'
            if tile_type_norm in ['LIOI_TBYTETERM', 'RIOI_TBYTETERM']:
                tile_type_norm = 'IOI'

            # ignore dummy tiles (ex: VBRK)
            if len(tiledata['bits']) == 0:
                if self.verbose:
                    for site in tiledata["sites"]:
                        assert site not in self.site_tags, "Site %s does not have bitstream info" % site
                this_tile_tags = len(self.tile_tags.get(tilename, {}))
                assert this_tile_tags == 0, "Tile %s does not have bitstream info but %s tags" % (
                    tilename, this_tile_tags)
                continue
            elif len(tiledata['bits']) == 1:
                bitj = list(tiledata['bits'].values())[0]
            else:
                assert self.def_bt in tiledata[
                    'bits'], 'Default block not present: %s' % self.def_bt
                bitj = tiledata['bits'][self.def_bt]

            # NOTE: multiple tiles may have the same base addr + offset
            segname = "%s_%03d" % (
                # truncate 0x to leave hex string
                bitj["baseaddr"][2:],
                bitj["offset"])

            # process tile name tags
            if tilename in self.tile_tags:
                add_tilename_tags()

            # process site name tags
            for site in tiledata["sites"]:
                if site not in self.site_tags:
                    continue
                add_site_tags()

        n_site_tags = recurse_sum(self.site_tags)
        n_tile_tags = recurse_sum(self.tile_tags)
        ntags = n_site_tags + n_tile_tags
        if self.verbose:
            assert ntags, "No tags"
            print("Used %u / %u tags" % (len(tags_used), ntags))
            print("Tag sites: %u" % (n_site_tags, ))
            if n_site_tags:
                print('  Ex: %s' % list(self.site_tags.keys())[0])
            print("Tag tiles: %u" % (n_tile_tags, ))
            print("Used %u sites" % len(sites_used))
            print("Grid DB had %u tile types" % len(tile_types_found))
        assert ntags == len(tags_used), "Unused tags, %s used out of %s" % (
            len(tags_used), ntags)

    def write(self, suffix=None, roi=False, allow_empty=False):
        assert self.segments_by_type, 'No data to write'

        if not allow_empty:
            assert sum(
                [len(segments) for segments in self.segments_by_type.values()
                 ]) != 0, "Didn't  generate any segments"

        for segtype in self.segments_by_type.keys():
            if suffix is not None:
                filename = "segdata_%s_%s.txt" % (segtype.lower(), suffix)
            else:
                filename = "segdata_%s.txt" % (segtype.lower())

            segments = self.segments_by_type[segtype]
            if segments:
                print("Writing %s." % filename)
                with OpenSafeFile(filename, "w") as f:
                    for segname, segdata in sorted(segments.items()):
                        # seg 00020300_010
                        print("seg %s" % segname, file=f)
                        for bitname in sorted(segdata["bits"]):
                            print("bit %s" % bitname, file=f)
                        for tagname, tagval in sorted(segdata["tags"].items()):
                            print("tag %s %d" % (tagname, tagval), file=f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Emits top.v's for various BUFHCE routing inputs. """
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.lut_maker import LutMaker
from prjxray.db import Database
from io import StringIO

CMT_XY_FUN = util.create_xy_fun(prefix='')
BUFGCTRL_XY_FUN = util.create_xy_fun('BUFGCTRL_')


def read_site_to_cmt():
    """ Yields clock sources and which CMT they route within. """
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt = l.strip().split(',')
            yield (site, cmt)


class ClockSources(object):
    """ Class for tracking clock sources.

    Some clock sources can be routed to any CMT, for these, cmt='ANY'.
    For clock sources that belong to a CMT, cmt should be set to the CMT of
    the source.

    """

    def __init__(self):
        self.sources = {}
        self.merged_sources = {}
        self.source_to_cmt = {}
        self.used_sources_from_cmt = {}

    def add_clock_source(self, source, cmt):
        """ Adds a source from a specific CMT.

        cmt='ANY' indicates that this source can be routed to any CMT.
        """
        if cmt not in self.sources:
            self.sources[cmt] = []

        self.sources[cmt].append(source)
        assert source not in self.source_to_cmt or self.source_to_cmt[
            source] == cmt, source
        self.source_to_cmt[source] = cmt

    def get_random_source(self, cmt):
        """ Get a random source that is routable to the specific CMT.

        get_random_source will return a source that is either cmt='ANY',
        cmt equal to the input CMT, or the adjecent CMT.

        """
        if cmt not in self.merged_sources:
            choices = []
            if 'ANY' in self.sources:
                choices.extend(self.sources['ANY'])

            if cmt in self.sources:
                choices.extend(self.sources[cmt])

            x, y = CMT_XY_FUN(cmt)

            if x % 2 == 0:
                x += 1
            else:
                x -= 1

            paired_cmt = 'X{}Y{}'.format(x, y)

            if paired_cmt in self.sources:
                choices.extend(self.sources[paired_cmt])

            self.merged_sources[cmt] = choices

        if self.merged_sources[cmt]:
            source = random.choice(self.merged_sources[cmt])

            source_cmt = self.source_to_cmt[source]
            if source_cmt not in self.used_sources_from_cmt:
                self.used_sources_from_cmt[source_cmt] = set()

            self.used_sources_from_cmt[source_cmt].add(source)

            if source_cmt != 'ANY' and len(
                    self.used_sources_from_cmt[source_cmt]) > 14:
                print('//', self.used_sources_from_cmt)
                self.used_sources_from_cmt[source_cmt].remove(source)
                return None
            else:
                return source


def main():
    """
    BUFG's can be driven from:

        Interconnect
        HROW cascade

    """

    print(
        '''
module top();
    (* KEEP, DONT_TOUCH *)
    LUT6 dummy();
    ''')

    site_to_cmt = dict(read_site_to_cmt())
    luts = LutMaker()
    wires = StringIO()
    bufgs = StringIO()

    clock_sources = ClockSources()

    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    def gen_sites(desired_site_type):
        for tile_name in sorted(grid.tiles()):
            loc = grid.loc_of_tilename(tile_name)
            gridinfo = grid.gridinfo_at_loc(loc)
            for site, site_type in gridinfo.sites.items():
                if site_type == desired_site_type:
                    yield tile_name, site

    for _, site in gen_sites('MMCME2_ADV'):
        mmcm_clocks = [
            'mmcm_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(13)
        ]

        for clk in mmcm_clocks:
            clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire {c0}, {c1}, {c2}, {c3}, {c4}, {c5};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    MMCME2_ADV pll_{site} (
        .CLKOUT0({c0}),
        .CLKOUT0B({c1}),
        .CLKOUT1({c2}),
        .CLKOUT1B({c3}),
        .CLKOUT2({c4}),
        .CLKOUT2B({c5}),
        .CLKOUT3({c6}),
        .CLKOUT3B({c7}),
        .CLKOUT4({c8}),
        .CLKOUT5({c9}),
        .CLKOUT6({c10}),
        .CLKFBOUT({c11}),
        .CLKFBOUTB({c12})
    );
        """.format(
                site=site,
                c0=mmcm_clocks[0],
                c1=mmcm_clocks[1],
                c2=mmcm_clocks[2],
                c3=mmcm_clocks[3],
                c4=mmcm_clocks[4],
                c5=mmcm_clocks[5],
                c6=mmcm_clocks[6],
                c7=mmcm_clocks[7],
                c8=mmcm_clocks[8],
                c9=mmcm_clocks[9],
                c10=mmcm_clocks[10],
                c11=mmcm_clocks[11],
                c12=mmcm_clocks[12],
            ))

    for _, site in sorted(gen_sites("BUFGCTRL"),
                          key=lambda x: BUFGCTRL_XY_FUN(x[1])):
        print(
            """
    wire O_{site};
    wire S1_{site};
    wire S0_{site};
    wire IGNORE1_{site};
    wire IGNORE0_{site};
    wire I1_{site};
    wire I0_{site};
    wire CE1_{site};
    wire CE0_{site};
    """.format(site=site),
            file=wires)

        print(
            """
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFGCTRL bufg_{site} (
        .O(O_{site}),
        .S1(S1_{site}),
        .S0(S0_{site}),
        .IGNORE1(IGNORE1_{site}),
        .IGNORE0(IGNORE0_{site}),
        .I1(I1_{site}),
        .I0(I0_{site}),
        .CE1(CE1_{site}),
        .CE0(CE0_{site})
        );
        """.format(site=site),
            file=bufgs)
    """ BUFG clock sources:

    2 from interconnect
    Output of BUFG +/- 1
    Cascade in (e.g. PLL, MMCM)

    """

    CLOCK_CHOICES = (
        'LUT',
        'BUFG_+1',
        'BUFG_-1',
        'CASCADE',
    )

    def find_bufg_cmt(tile):
        if '_BOT_' in tile:
            inc = 1
        else:
            inc = -1

        loc = grid.loc_of_tilename(tile)

        offset = 1

        while True:
            gridinfo = grid.gridinfo_at_loc(
                (loc.grid_x, loc.grid_y + offset * inc))
            if gridinfo.tile_type.startswith('CLK_HROW_'):
                return site_to_cmt[list(gridinfo.sites.keys())[0]]

            offset += 1

    def get_clock_net(tile, site, source_type):
        if source_type == 'LUT':
            return luts.get_next_output_net()
        elif source_type == 'BUFG_+1':
            x, y = BUFGCTRL_XY_FUN(site)

            target_y = y + 1
            max_y = ((y // 16) + 1) * 16

            if target_y >= max_y:
                target_y -= 16

            return 'O_BUFGCTRL_X{x}Y{y}'.format(x=x, y=target_y)
        elif source_type == 'BUFG_-1':
            x, y = BUFGCTRL_XY_FUN(site)

            target_y = y - 1
            min_y = (y // 16) * 16

            if target_y < min_y:
                target_y += 16

            return 'O_BUFGCTRL_X{x}Y{y}'.format(x=x, y=target_y)
        elif source_type == 'CASCADE':
            cmt = find_bufg_cmt(tile)
            return clock_sources.get_random_source(cmt)
        else:
            assert False, source_type

    for tile, site in sorted(gen_sites("BUFGCTRL"),
                             key=lambda x: BUFGCTRL_XY_FUN(x[1])):
        if random.randint(0, 1):
            print(
                """
    assign I0_{site} = {i0_net};""".format(
                    site=site,
                    i0_net=get_clock_net(
                        tile, site, random.choice(CLOCK_CHOICES))),
                file=bufgs)

        if random.randint(0, 1):
            print(
                """
    assign I1_{site} = {i1_net};""".format(
                    site=site,
                    i1_net=get_clock_net(
                        tile, site, random.choice(CLOCK_CHOICES))),
                file=bufgs)

        print(
            """
    assign S0_{site} = {s0_net};
    assign S1_{site} = {s1_net};
    assign IGNORE0_{site} = {ignore0_net};
    assign IGNORE1_{site} = {ignore1_net};
    assign CE0_{site} = {ce0_net};
    assign CE1_{site} = {ce1_net};
        """.format(
                site=site,
                s0_net=luts.get_next_output_net(),
                s1_net=luts.get_next_output_net(),
                ignore0_net=luts.get_next_output_net(),
                ignore1_net=luts.get_next_output_net(),
                ce0_net=luts.get_next_output_net(),
                ce1_net=luts.get_next_output_net(),
            ),
            file=bufgs)

    for l in luts.create_wires_and_luts():
        print(l)

    print(wires.getvalue())
    print(bufgs.getvalue())

    itr = iter(gen_sites('BUFHCE'))

    for tile, site in sorted(gen_sites("BUFGCTRL"),
                             key=lambda x: BUFGCTRL_XY_FUN(x[1])):
        if random.randint(0, 1):
            _, bufhce_site = next(itr)

            print(
                """
    (* KEEP, DONT_TOUCH, LOC = "{bufhce_site}" *)
    BUFHCE bufhce_{bufhce_site} (
        .I(O_{site})
        );""".format(
                    site=site,
                    bufhce_site=bufhce_site,
                ))

    print("endmodule")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This script allows to find missing segbits in the database.

For each tile the script loads its 'tile_type_*.json' file and looks for all
non-pseudo pips there. Next it loads corresponding 'segbits_*.db' file (if found)
and checks if those pips are listed there.

Missing segbits for pips are reported as well as missing segbit files.
"""
import sys
import logging
import json
import argparse
import os
import re

from prjxray.util import OpenSafeFile

# =============================================================================


def read_pips_from_tile(tile_file):
    """
    Loads pip definition from a tile type JSON file and returns non-pseudo
    PIP name strings. Names are formatted as <dst_wire>.<src_wire>
    """

    with OpenSafeFile(tile_file, "r") as fp:
        root = json.load(fp)
        pips = root["pips"]

        pip_names = []
        for pip in pips.values():
            if int(pip["is_pseudo"]) == 0:
                pip_names.append(
                    "{}.{}".format(pip["dst_wire"], pip["src_wire"]))

    return pip_names


def read_ppips(ppips_file):
    """
    Loads and parses ppips_*.db file. Returns a dict indexed by PIP name which
    contains their types ("always", "default" or "hint")
    """
    ppips = {}

    with OpenSafeFile(ppips_file, "r") as fp:
        for line in fp.readlines():
            line = line.split()
            if len(line) == 2:
                full_pip_name = line[0].split(".")
                pip_name = ".".join(full_pip_name[1:])
                ppips[pip_name] = line[1]

    return ppips


def read_segbits(segbits_file):
    """
    Loads and parses segbits_*.db file. Returns only segbit names.
    """
    segbits = []

    with OpenSafeFile(segbits_file, "r") as fp:
        for line in fp.readlines():
            line = line.split()
            if len(line) > 1:
                fields = line[0].split(".")
                segbit = ".".join(fields[1:])
                segbits.append(segbit)

    return segbits


# =============================================================================


def main(argv):
    """
    The main
    """

    exitcode = 0

    # Parse arguments
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "--db-root", type=str, required=True, help="Database root")
    parser.add_argument(
        "--verbose", type=int, default=0, help="Verbosity level 0-5")
    parser.add_argument(
        "--skip-tiles",
        type=str,
        nargs="*",
        default=[],
        help="Tile type name regex list for tile types to skip")
    parser.add_argument(
        "--incl-tiles",
        type=str,
        nargs="*",
        default=[],
        help="Tile type name regex list for tile types to include")

    args = parser.parse_args(argv[1:])

    logging.basicConfig(level=50 - args.verbose * 10, format="%(message)s")

    # List files in DB root
    files = os.listdir(args.db_root)

    # List tile types
    tile_types = []
    for file in files:
        match = re.match("^tile_type_(\\w+).json$", file)
        if match:
            tile_types.append(match.group(1))

    tile_types.sort()

    # Look for missing bits
    for tile_type in tile_types:

        # Check if we should include this tile
        do_skip = len(args.incl_tiles) > 0
        for pattern in args.incl_tiles:
            if re.match(pattern, tile_type):
                do_skip = False
                break

        # Check if we should skip this tile
        for pattern in args.skip_tiles:
            if re.match(pattern, tile_type):
                do_skip = True
                break

        if do_skip:
            continue

        logging.critical(tile_type)

        # DB file names
        tile_file = os.path.join(
            args.db_root, "tile_type_{}.json".format(tile_type.upper()))
        ppips_file = os.path.join(
            args.db_root, "ppips_{}.db".format(tile_type.lower()))
        segbits_file = os.path.join(
            args.db_root, "segbits_{}.db".format(tile_type.lower()))

        # Load pips
        pips = read_pips_from_tile(tile_file)

        # Load ppips (if any)
        if os.path.isfile(ppips_file):
            ppips = read_ppips(ppips_file)
        else:
            ppips = {}

        # Load segbits (if any)
        if os.path.isfile(segbits_file):
            segbits = read_segbits(segbits_file)
        else:
            segbits = []

        # There are non-pseudo pips in this tile
        if len(pips):
            missing_bits = 0
            known_bits = 0

            # Build a list of pips to check. If a pip is listed in the ppips
            # file and it is not "default" then make it a pseudo one
            pips_to_check = []
            for pip in pips:
                if pip in ppips.keys() and ppips[pip] != "default":
                    continue
                pips_to_check.append(pip)

            # Missing segbits file
            if len(segbits) == 0:
                missing_bits = len(pips_to_check)
                logging.critical(" MISSING: no segbits file!")
                exitcode = -1

            # Segbits file present
            else:

                # Check pips
                for pip in pips_to_check:
                    if pip not in segbits:

                        # A "default" pip
                        if pip in ppips.keys() and ppips[pip] == "default":
                            missing_bits += 1
                            logging.error(
                                " WARNING: no bits for pip '{}' which defaults to VCC_WIRE"
                                .format(pip))
                            exitcode = -1

                        # A regular pip
                        else:
                            missing_bits += 1
                            logging.error(
                                " MISSING: no bits for pip '{}'".format(pip))
                            exitcode = -1

                    # The pip has segbits
                    else:
                        known_bits += 1

            # Report missing bit count
            if missing_bits > 0:
                logging.critical(
                    " MISSING: no bits for {}/{} pips!".format(
                        missing_bits, missing_bits + known_bits))
                exitcode = -1
            else:
                logging.critical(" OK: no missing bits")

        # No pips
        else:
            logging.warning(" OK: no pips")

    return exitcode


# =============================================================================

if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
from prjxray import util

segmk = Segmaker("design.bits")
cache = dict()

print("Loading tags")
'''
module,loc,n
clb_NFFMUX_O5,SLICE_X12Y100,3
clb_NFFMUX_AX,SLICE_X13Y100,2
clb_NFFMUX_O6,SLICE_X14Y100,3
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    module, loc, n = l.split(',')
    n = int(n)
    which = chr(ord('A') + n)
    # clb_NFFMUX_AX => AX
    src = module.replace('clb_NOUTMUX_', '')
    '''
    BOUTMUX
            30_20   30_21    30_22   30_23
    O6      1
    O5      1               1
    XOR             1
    CY              1       1
    F8                      1       1
    B5Q                             1
    '''

    # if location not included in cache yet: start with assuming all four MUXes are unused.
    if loc not in cache:
        cache[loc] = set("ABCD")

    # rewrite name of F78 source net: MUXes A and C have an F7 input, MUX B has an F8 input
    if src == "F78":
        if which in "AC":
            src = "F7"
        elif which == "B":
            src = "F8"
        else:
            assert 0

    # rewrite name of B5Q source net: It's actually A5Q, B5Q, C5Q, or D5Q
    if src == "B5Q":
        src = which + "5Q"

    # add the 1-tag for this connection
    tag = "%sOUTMUX.%s" % (which, src)
    segmk.add_site_tag(loc, tag, 1)

    # remove this MUX from the cache, preventing generation of 0-tags for this MUX
    cache[loc].remove(which)

    # O6 hack per https://github.com/SymbiFlow/prjxray/issues/243
    segmk.add_site_tag(loc, "%sOUTMUX.%s" % (which, "O6"), src == "O5")

# create 0-tags for all sources on the remaining (unused) MUXes
for loc, muxes in cache.items():
    for which in muxes:
        for src in "F7 F8 CY O5 XOR 5Q MC31".split():
            if src == "MC31" and which is not "D": continue
            if src == "F7" and which not in "AC": continue
            if src == "F8" and which not in "B": continue
            if src == "5Q": src = which + "5Q"
            tag = "%sOUTMUX.%s" % (which, src)
            segmk.add_site_tag(loc, tag, 0)


def bitfilter(frame_idx, bit_idx):
    # locations of A5MA, B5MA, C5MA, D5MA bits. because of the way we generate specimens
    # in this fuzzer we get some aliasing with those bits, so we have to manually exclude
    # them. (Maybe FIXME: read the bit locations from the database files)

    # Since the SRL32 is enabled along with DOUTMUX.MC31, bits related to
    # SRL32 features are masked out.

    if (frame_idx, bit_idx) in [
        (30, 55),
        (31, 55),  # D5MA
        (31, 44),
        (31, 45),  # C5MA
        (30, 19),
        (31, 19),  # B5MA
        (30, 9),
        (31, 8),  # A5MA
        (30, 16),  # ALUT.SRL
        (1, 23),  # WEMUX.CE
    ]:
        return False

    return util.bitfilter_clb_mux(frame_idx, bit_idx)


segmk.compile(bitfilter=bitfilter)
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import fasm
import sys

from prjxray import bitstream


class FasmLookupError(Exception):
    pass


class FasmInconsistentBits(Exception):
    pass


def init_frame_at_address(frames, addr):
    '''Set given frame to 0 if not initialized '''
    if not addr in frames:
        frames[addr] = [0 for _i in range(bitstream.FRAME_WORD_COUNT)]


class FasmAssembler(object):
    def __init__(self, db):
        self.db = db
        self.grid = db.grid()

        self.seen_tile = set()
        self.frames_in_use = set()

        self.frames = {}
        self.frames_line = {}

        self.feature_callback = lambda feature: None

    def set_feature_callback(self, feature_callback):
        self.feature_callback = feature_callback

    def get_frames(self, sparse=False):
        if not sparse:
            frames = self.frames_init()
        else:
            # Even in sparse mode, zero all frames for any tile that is
            # setting a bit.  This handles the case where the tile has
            # multiple frames, but the FASM only specifies some of the frames.
            frames = {}
            for frame in self.frames_in_use:
                init_frame_at_address(frames, frame)

        for (frame_addr, word_addr, bit_index), is_set in self.frames.items():
            init_frame_at_address(frames, frame_addr)

            if word_addr >= 101:
                print(f"get_frames: invalid word address {word_addr} in line: {self.frames_line[(frame_addr, word_addr, bit_index)]}", file=sys.stderr)
            if frame_addr not in frames:
                print(f"get_frames: invalid frame address {frame_addr:x8}", file=sys.stderr)
            if is_set:
                frames[frame_addr][word_addr] |= 1 << bit_index

        return frames

    def frames_init(self):
        '''Set all frames to 0'''
        frames = {}

        for bits_info in self.grid.iter_all_frames():
            for coli in range(bits_info.bits.frames):
                init_frame_at_address(
                    frames, bits_info.bits.base_address + coli)

        return frames

    def frame_set(self, frame_addr, word_addr, bit_index, line):
        '''Set given bit in given frame address and word'''
        assert bit_index is not None

        key = (frame_addr, word_addr, bit_index)
        if word_addr >= 101:
            print(f"frame_set: invalid word address {word_addr} in line: {line}", file=sys.stderr)
            return
        if key in self.frames:
            if self.frames[key] != 1:
                raise FasmInconsistentBits(
                    'FASM line "{}" wanted to set bit {} but was cleared by FASM line "{}"'
                    .format(
                        line,
                        key,
                        self.frames_line[key],
                    ))
            return

        self.frames[key] = 1
        self.frames_line[key] = line

    def frame_clear(self, frame_addr, word_addr, bit_index, line):
        '''Set given bit in given frame address and word'''
        assert bit_index is not None

        key = (frame_addr, word_addr, bit_index)
        if word_addr >= 101:
            print(f"frame_clear: invalid word address {word_addr} in line: {line}", file=sys.stderr)
            return
        if key in self.frames:
            if self.frames[key] != 0:
                raise FasmInconsistentBits(
                    'FASM line "{}" wanted to clear bit {} but was set by FASM line "{}"'
                    .format(
                        line,
                        key,
                        self.frames_line[key],
                    ))
            return

        self.frames[key] = 0
        self.frames_line[key] = line

    def enable_feature(self, tile, feature, address, line):
        gridinfo = self.grid.gridinfo_at_tilename(tile)

        def update_segbit(bit):
            '''Set or clear a single bit in a segment at the given word column and word bit position'''

            frame_addr = bit.word_column
            word_addr = bit.word_bit // bitstream.WORD_SIZE_BITS
            bit_index = bit.word_bit % bitstream.WORD_SIZE_BITS

            if bit.isset:
                self.frame_set(frame_addr, word_addr, bit_index, line)
            else:
                self.frame_clear(frame_addr, word_addr, bit_index, line)

        segbits = self.grid.get_tile_segbits_at_tilename(tile)

        self.seen_tile.add(tile)

        db_k = '%s.%s' % (gridinfo.tile_type, feature)

        any_bits = set()

        try:
            for block_type, bit in segbits.feature_to_bits(gridinfo.bits, db_k,
                                                           address):
                any_bits.add(block_type)
                update_segbit(bit)
        except KeyError:
            raise FasmLookupError(
                "Segment DB %s, key %s not found from line '%s'" %
                (gridinfo.tile_type, db_k, line))

        for block_type in any_bits:
            # Mark all frames used by this tile as in use.
            bits = gridinfo.bits[block_type]
            for frame in range(bits.base_address,
                               bits.base_address + bits.frames):
                self.frames_in_use.add(frame)

    def add_fasm_line(self, line, missing_features):
        if not line.set_feature:
            return

        self.feature_callback(line.set_feature)

        line_strs = tuple(fasm.fasm_line_to_string(line))
        assert len(line_strs) == 1
        line_str = line_strs[0]

        parts = line.set_feature.feature.split('.')
        tile = parts[0]
        feature = '.'.join(parts[1:])

        # canonical_features flattens multibit feature enables to only
        # single bit features, which is what enable_feature expects.
        #
        # canonical_features also filters out features that are not enabled,
        # which are no-ops.
        for flat_set_feature in fasm.canonical_features(line.set_feature):
            address = 0
            if flat_set_feature.start is not None:
                address = flat_set_feature.start

            try:
                self.enable_feature(tile, feature, address, line_str)
            except FasmLookupError as e:
                missing_features.append(str(e))

    def parse_fasm_filename(self, filename, extra_features=[]):
        missing_features = []
        for line in fasm.parse_fasm_filename(filename):
            self.add_fasm_line(line, missing_features)

        for line in extra_features:
            self.add_fasm_line(line, missing_features)

        if missing_features:
            raise FasmLookupError('\n'.join(missing_features))

    def mark_roi_frames(self, roi):
        for tile in roi.gen_tiles():
            gridinfo = self.grid.gridinfo_at_tilename(tile)

            for block_type in gridinfo.bits:
                bits = gridinfo.bits[block_type]
                for frame in range(bits.base_address,
                                   bits.base_address + bits.frames):
                    self.frames_in_use.add(frame)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.verilog import vrandbit, vrandbits
import sys
import json


def gen_sites():
    for _tile_name, site_name, _site_type in sorted(util.get_roi().gen_sites(
        ["XADC"])):
        yield site_name


sites = list(gen_sites())
DUTN = len(sites)
DIN_N = DUTN * 8
DOUT_N = DUTN * 8

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.jl', 'w')
f.write('module,loc,params\n')
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))

for loci, site in enumerate(sites):

    ports = {
        'clk': 'clk',
        'din': 'din[  %d +: 8]' % (8 * loci, ),
        'dout': 'dout[  %d +: 8]' % (8 * loci, ),
    }

    params = {
        "INIT_43": random.randint(0x000, 0xFFFF),
    }

    modname = "my_XADC"
    verilog.instance(modname, "inst_%u" % loci, ports, params=params)
    # LOC isn't support
    params["LOC"] = verilog.quote(site)

    j = {'module': modname, 'i': loci, 'params': params}
    f.write('%s\n' % (json.dumps(j)))
    print('')

f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module my_XADC (input clk, input [7:0] din, output [7:0] dout);
    parameter INIT_43 = 16'h0000;

    (* KEEP, DONT_TOUCH *)
    XADC #(
        .INIT_43(INIT_43)
        ) dut (
            .BUSY(),
            .DRDY(),
            .EOC(),
            .EOS(),
            .JTAGBUSY(),
            .JTAGLOCKED(),
            .JTAGMODIFIED(),
            .OT(),
            .DO(),
            .ALM(),
            .CHANNEL(),
            .MUXADDR(),
            .CONVST(),
            .CONVSTCLK(clk),
            .DCLK(clk),
            .DEN(),
            .DWE(),
            .RESET(),
            .VN(),
            .VP(),
            .DI(),
            .VAUXN(),
            .VAUXP(),
            .DADDR());
endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Emits top.v's for various BUFHCE routing inputs. """
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.lut_maker import LutMaker
from prjxray.db import Database
from io import StringIO

CMT_XY_FUN = util.create_xy_fun(prefix='')


def read_site_to_cmt():
    """ Yields clock sources and which CMT they route within. """
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt = l.strip().split(',')
            yield (site, cmt)


class ClockSources(object):
    """ Class for tracking clock sources.
    """

    def __init__(self, limit=14):
        self.sources = {}
        self.source_to_cmt = {}
        self.used_sources_from_cmt = {}
        self.limit = limit

    def add_clock_source(self, source, cmt):
        """ Adds a source from a specific CMT.

        cmt='ANY' indicates that this source can be routed to any CMT.
        """
        if cmt not in self.sources:
            self.sources[cmt] = []

        self.sources[cmt].append(source)
        self.source_to_cmt[source] = cmt

    def get_random_source(self, cmt, no_repeats=False):
        """ Get a random source that is routable to the specific CMT.

        get_random_source will return a source that is either cmt='ANY',
        cmt equal to the input CMT, or the adjecent CMT.

        """

        choices = []

        if cmt in self.sources:
            choices.extend(self.sources[cmt])

        random.shuffle(choices)
        for source in choices:

            source_cmt = self.source_to_cmt[source]

            if source_cmt not in self.used_sources_from_cmt:
                self.used_sources_from_cmt[source_cmt] = set()

            if no_repeats and source in self.used_sources_from_cmt[source_cmt]:
                continue

            if len(self.used_sources_from_cmt[source_cmt]) >= self.limit:
                continue

            self.used_sources_from_cmt[source_cmt].add(source)
            return source

        return None


def main():
    """
    HCLK_IOI has the following inputs:

    12 (east) BUFH from the right side of the HROW
    12 (west) Bounce PIPs from one BUFH to any of 6 GCLK_BOT and 6 GCLK_TOP
    4 (east) PHSR_PERFCLK (IOCLK_PLL) from HCLK_CLB to input of BUFIO
    8 (4 north and 4 south) BUFR CLR and CE
    2 (south) I2IOCLK to input of BUFR
    2 (north) I2IOCLK to input of BUFR
    2 RCLK IMUX (IMUX0 and IMUX1) choosing input of BUFR

    outputs:
    4 (east) BUFRCLK - from BUFR to HROW
    4 (north) BUFR2IO - from BUFR
    4 (north) IOCLK from BUFIO

    """

    global_clock_sources = ClockSources()
    cmt_clock_sources = ClockSources()
    cmt_fast_clock_sources = ClockSources(4)
    bufr_clock_sources = ClockSources()
    bufio_clock_sources = ClockSources()
    site_to_cmt = dict(read_site_to_cmt())
    clock_region_limit = dict()
    clock_region_serdes_location = dict()

    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    def gen_sites(desired_site_type):
        for tile_name in sorted(grid.tiles()):
            loc = grid.loc_of_tilename(tile_name)
            gridinfo = grid.gridinfo_at_loc(loc)
            for site, site_type in gridinfo.sites.items():
                if site_type == desired_site_type:
                    yield tile_name, site

    def serdes_relative_location(tile, site):
        (serdes_loc_x, serdes_loc_y) = grid.loc_of_tilename(tile)
        serdes_clk_reg = site_to_cmt[site]
        for tile_name in sorted(grid.tiles()):
            if 'HCLK_IOI3' in tile_name:
                (hclk_tile_loc_x,
                 hclk_tile_loc_y) = grid.loc_of_tilename(tile_name)
                if hclk_tile_loc_x == serdes_loc_x:
                    gridinfo = grid.gridinfo_at_loc(
                        (hclk_tile_loc_x, hclk_tile_loc_y))
                    random_site = next(iter(gridinfo.sites.keys()))
                    hclk_clk_reg = site_to_cmt[random_site]
                    if hclk_clk_reg == serdes_clk_reg:
                        if serdes_loc_y < hclk_tile_loc_y:
                            return "TOP"
                        elif serdes_loc_y > hclk_tile_loc_y:
                            return "BOTTOM"
                        else:
                            assert False

    clock_region_sites = set()

    def get_clock_region_site(site_type, clk_reg):
        for site_name, reg in site_to_cmt.items():
            if site_name.startswith(site_type) and reg in clk_reg:
                if site_name not in clock_region_sites:
                    clock_region_sites.add(site_name)
                    return site_name

    print(
        '''
module top();
    (* KEEP, DONT_TOUCH *)
    LUT6 dummy();
    ''')

    luts = LutMaker()
    bufs = StringIO()

    for _, site in gen_sites('MMCME2_ADV'):
        mmcm_clocks = [
            'mmcm_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(13)
        ]

        for idx, clk in enumerate(mmcm_clocks):
            if idx < 4:
                cmt_fast_clock_sources.add_clock_source(clk, site_to_cmt[site])
            else:
                cmt_clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire cin1_{site}, cin2_{site}, clkfbin_{site}, {c0}, {c1}, {c2}, {c3}, {c4}, {c5};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    MMCME2_ADV pll_{site} (
        .CLKIN1(cin1_{site}),
        .CLKIN2(cin2_{site}),
        .CLKFBIN(clkfbin_{site}),
        .CLKOUT0({c0}),
        .CLKOUT0B({c4}),
        .CLKOUT1({c1}),
        .CLKOUT1B({c5}),
        .CLKOUT2({c2}),
        .CLKOUT2B({c6}),
        .CLKOUT3({c3}),
        .CLKOUT3B({c7}),
        .CLKOUT4({c8}),
        .CLKOUT5({c9}),
        .CLKOUT6({c10}),
        .CLKFBOUT({c11}),
        .CLKFBOUTB({c12})
    );
        """.format(
                site=site,
                c0=mmcm_clocks[0],
                c1=mmcm_clocks[1],
                c2=mmcm_clocks[2],
                c3=mmcm_clocks[3],
                c4=mmcm_clocks[4],
                c5=mmcm_clocks[5],
                c6=mmcm_clocks[6],
                c7=mmcm_clocks[7],
                c8=mmcm_clocks[8],
                c9=mmcm_clocks[9],
                c10=mmcm_clocks[10],
                c11=mmcm_clocks[11],
                c12=mmcm_clocks[12],
            ))

    for _, site in gen_sites('PLLE2_ADV'):
        pll_clocks = [
            'pll_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(7)
        ]

        for clk in pll_clocks:
            cmt_clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire cin1_{site}, cin2_{site}, clkfbin_{site}, {c0}, {c1}, {c2}, {c3}, {c4}, {c5}, {c6};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    PLLE2_ADV pll_{site} (
        .CLKIN1(cin1_{site}),
        .CLKIN2(cin2_{site}),
        .CLKFBIN(clkfbin_{site}),
        .CLKOUT0({c0}),
        .CLKOUT1({c1}),
        .CLKOUT2({c2}),
        .CLKOUT3({c3}),
        .CLKOUT4({c4}),
        .CLKOUT5({c5}),
        .CLKFBOUT({c6})
    );
        """.format(
                site=site,
                c0=pll_clocks[0],
                c1=pll_clocks[1],
                c2=pll_clocks[2],
                c3=pll_clocks[3],
                c4=pll_clocks[4],
                c5=pll_clocks[5],
                c6=pll_clocks[6],
            ))

    for tile_name, site in gen_sites('BUFHCE'):
        print(
            """
    wire I_{site};
    wire O_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFHCE buf_{site} (
        .I(I_{site}),
        .O(O_{site})
    );""".format(site=site),
            file=bufs)
        global_clock_sources.add_clock_source(
            'O_{site}'.format(site=site), site_to_cmt[site])

    hclks_used_by_clock_region = {}
    for cmt in site_to_cmt.values():
        hclks_used_by_clock_region[cmt] = set()

    def check_hclk_src(src, src_cmt):
        if len(hclks_used_by_clock_region[src_cmt]
               ) >= 12 and src not in hclks_used_by_clock_region[src_cmt]:
            return None
        else:
            hclks_used_by_clock_region[src_cmt].add(src)
            return src

    cmt_clks_used_by_clock_region = {}
    for cmt in site_to_cmt.values():
        cmt_clks_used_by_clock_region[cmt] = list()

    def check_cmt_clk_src(src, src_clock_region):
        print(
            "//src: {}, clk_reg: {}, len {}".format(
                src, src_clock_region,
                len(cmt_clks_used_by_clock_region[src_clock_region])))
        if len(cmt_clks_used_by_clock_region[src_clock_region]) >= 4:
            return None
        else:
            cmt_clks_used_by_clock_region[src_clock_region].append(src)
            return src

    #Add IDELAYCTRL
    idelayctrl_in_clock_region = {}
    for cmt in site_to_cmt.values():
        idelayctrl_in_clock_region[cmt] = False
    for _, site in gen_sites('IDELAYCTRL'):
        if random.random() < 0.5:
            wire_name = global_clock_sources.get_random_source(
                site_to_cmt[site], no_repeats=False)
            if wire_name is None:
                continue
            src_cmt = global_clock_sources.source_to_cmt[wire_name]
            wire_name = check_hclk_src(wire_name, src_cmt)

            if wire_name is None:
                continue
            idelayctrl_in_clock_region[src_cmt] = True
            print(
                """
        assign I_{site} = {clock_source};
        (* KEEP, DONT_TOUCH, LOC = "{site}" *)
        IDELAYCTRL idelay_ctrl_{site} (
            .RDY(),
            .REFCLK(I_{site}),
            .RST()
            );""".format(site=site, clock_source=wire_name))

    # Add SERDES driven by BUFH or MMCM
    for tile, site in gen_sites('ILOGICE3'):
        wire_name = None
        clock_region = site_to_cmt[site]
        if clock_region not in clock_region_limit:
            # Select serdes limit and relative location per clock region
            serdes_location = random.choice(["TOP", "BOTTOM", "ANY"])
            if serdes_location in "ANY":
                #We want TOP and BOTTOM IGCLK PIPs occupied but leave one slot for IDELAYCTRL
                if idelayctrl_in_clock_region[clock_region]:
                    clock_region_limit[clock_region] = 0 if random.random(
                    ) < 0.2 else 11
                else:
                    clock_region_limit[clock_region] = 0 if random.random(
                    ) < 0.2 else 12
            else:
                if idelayctrl_in_clock_region[clock_region]:
                    clock_region_limit[clock_region] = 0 if random.random(
                    ) < 0.2 else 5
                else:
                    clock_region_limit[clock_region] = 0 if random.random(
                    ) < 0.2 else 6

            clock_region_serdes_location[clock_region] = serdes_location

        # We reached the limit of hclks in this clock region
        if clock_region_limit[clock_region] == 0:
            continue

        # Add a serdes if it's located at the correct side from the HCLK_IOI tile
        if clock_region_serdes_location[clock_region] not in "ANY" and \
                serdes_relative_location(tile, site) != clock_region_serdes_location[clock_region]:
            continue
        if random.random() > 0.1:
            wire_name = global_clock_sources.get_random_source(
                site_to_cmt[site], no_repeats=True)
            if wire_name is None:
                continue
            src_cmt = global_clock_sources.source_to_cmt[wire_name]
            wire_name = check_hclk_src(wire_name, src_cmt)
            if wire_name is None:
                print("//wire is None")
                continue
            clock_region_limit[clock_region] -= 1
            print(
                """
    assign serdes_clk_{site} = {clock_source};""".format(
                    site=site, clock_source=wire_name))
        else:
            wire_name = cmt_fast_clock_sources.get_random_source(
                site_to_cmt[site], no_repeats=False)
            if wire_name is None:
                continue
            src_cmt = cmt_fast_clock_sources.source_to_cmt[wire_name]
            wire_name = check_cmt_clk_src(wire_name, src_cmt)
            if wire_name is None:
                continue
            bufio_site = get_clock_region_site("BUFIO", clock_region)
            if bufio_site is None:
                continue
            print(
                """
    assign serdes_clk_{serdes_loc} = O_{site};
    assign I_{site} = {clock_source};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFIO bufio_{site} (
        .O(O_{site}),
        .I(I_{site})
    );""".format(site=bufio_site, clock_source=wire_name, serdes_loc=site))

        print(
            "// clock_region: {} {}".format(
                clock_region, clock_region_serdes_location[clock_region]))
        print(
            """
    (* KEEP, DONT_TOUCH, LOC = "{loc}" *)
    ISERDESE2 #(
    .DATA_RATE("SDR"),
    .DATA_WIDTH(4),
    .DYN_CLKDIV_INV_EN("FALSE"),
    .DYN_CLK_INV_EN("FALSE"),
    .INIT_Q1(1'b0),
    .INIT_Q2(1'b0),
    .INIT_Q3(1'b0),
    .INIT_Q4(1'b0),
    .INTERFACE_TYPE("OVERSAMPLE"),
    .IOBDELAY("NONE"),
    .NUM_CE(2),
    .OFB_USED("FALSE"),
    .SERDES_MODE("MASTER"),
    .SRVAL_Q1(1'b0),
    .SRVAL_Q2(1'b0),
    .SRVAL_Q3(1'b0),
    .SRVAL_Q4(1'b0)
    )
    ISERDESE2_inst_{loc} (
    .CLK(serdes_clk_{loc}),
    .CLKB(),
    .CLKDIV(),
    .D(1'b0),
    .DDLY(),
    .OFB(),
    .OCLKB(),
    .RST(),
    .SHIFTIN1(),
    .SHIFTIN2()
    );
    """.format(loc=site, clock_source=wire_name))

    # BUFRs
    for _, site in gen_sites('BUFR'):
        if random.random() < 0.6:
            if random.random() < 0.5:
                wire_name = luts.get_next_output_net()
            else:
                wire_name = cmt_fast_clock_sources.get_random_source(
                    site_to_cmt[site], no_repeats=False)
                if wire_name is None:
                    continue
                src_cmt = cmt_fast_clock_sources.source_to_cmt[wire_name]
                wire_name = check_cmt_clk_src(wire_name, src_cmt)
                if wire_name is None:
                    continue
            bufr_clock_sources.add_clock_source(
                'O_{site}'.format(site=site), site_to_cmt[site])

            # Add DIVIDE
            divide = "BYPASS"
            if random.random() < 0.5:
                divide = "".format(random.randint(2, 8))

            print(
                """
    assign I_{site} = {clock_source};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFR #(.BUFR_DIVIDE("{divide}")) bufr_{site} (
        .O(O_{site}),
        .I(I_{site})
        );""".format(site=site, clock_source=wire_name, divide=divide),
                file=bufs)

    for _, site in gen_sites('MMCME2_ADV'):
        wire_name = bufr_clock_sources.get_random_source(
            site_to_cmt[site], no_repeats=True)

        if wire_name is None:
            continue
        print(
            """
    assign cin1_{site} = {wire_name};""".format(
                site=site, wire_name=wire_name))

    print(bufs.getvalue())

    for l in luts.create_wires_and_luts():
        print(l)

    print("endmodule")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Check:
-Individual files are valid
-No overlap between any tile

TODO:
Can we use prjxray?
Relies on 074, which is too far into the process
'''

from prjxray import util
from prjxray import db as prjxraydb
import os
import utils.parsedb as parsedb
#from prjxray import db as prjxraydb
import glob


def gen_tile_bits(tile_segbits, tile_bits):
    '''
    For given tile and corresponding db_file structure yield
    (absolute address, absolute FDRI bit offset, tag)

    For each tag bit in the corresponding block_type entry, calculate absolute address and bit offsets
    '''

    for block_type in tile_segbits:
        assert block_type in tile_bits, "block type %s is not present in current tile" % block_type

        block = tile_bits[block_type]

        baseaddr = block.base_address
        bitbase = 32 * block.offset
        frames = block.frames

        for tag in tile_segbits[block_type]:
            for bit in tile_segbits[block_type][tag]:
                # 31_06
                word_column = bit.word_column
                word_bit = bit.word_bit
                assert word_column <= frames, "ERROR: bit out of bound --> tag: %s; word_column = %s; frames = %s" % (
                    tag, word_column, frames)
                yield word_column + baseaddr, word_bit + bitbase, tag


def make_tile_mask(tile_segbits, tile_name, tile_bits):
    '''
    Return dict
    key: (address, bit index)
    val: sample description of where it came from (there may be multiple, only one)
    '''

    # FIXME: fix mask files https://github.com/SymbiFlow/prjxray/issues/301
    # in the meantime build them on the fly
    # We may want this to build them anyway

    ret = dict()
    for absaddr, bitaddr, tag in gen_tile_bits(tile_segbits, tile_bits):
        name = "%s.%s" % (tile_name, tag)
        ret.setdefault((absaddr, bitaddr), name)
    return ret


def parsedb_all(db_root, verbose=False):
    '''Verify .db files are individually valid'''

    files = 0
    for bit_fn in glob.glob('%s/segbits_*.db' % db_root):
        # Don't parse db files with fuzzer origin information
        if "origin_info" in bit_fn:
            continue
        verbose and print("Checking %s" % bit_fn)
        parsedb.run(bit_fn, fnout=None, strict=True, verbose=verbose)
        files += 1
    print("segbits_*.db: %d okay" % files)

    files = 0
    for bit_fn in glob.glob('%s/mask_*.db' % db_root):
        verbose and print("Checking %s" % bit_fn)
        parsedb.run(bit_fn, fnout=None, strict=True, verbose=verbose)
        files += 1
    print("mask_*.db: %d okay" % files)


def check_tile_overlap(db, verbose=False):
    '''
    Verifies that no two tiles use the same bit

    Assume .db files are individually valid
    Create a mask for all the bits the tile type uses
    For each tile, create bitmasks over the entire bitstream for current part
    Throw an exception if two tiles share an address
    '''
    mall = dict()
    tiles_type_done = dict()
    tile_segbits = dict()
    grid = db.grid()
    tiles_checked = 0

    for tile_name in grid.tiles():
        tile_info = grid.gridinfo_at_tilename(tile_name)
        tile_type = tile_info.tile_type
        tile_bits = tile_info.bits

        if tile_type not in tiles_type_done:
            segbits = db.get_tile_segbits(tile_type).segbits
            tile_segbits[tile_type] = segbits

            # If segbits has zero length the tile_type is marked True in order to be skipped
            if len(segbits) == 0:
                tiles_type_done[tile_type] = True
            else:
                tiles_type_done[tile_type] = False

        if tiles_type_done[tile_type]:
            continue

        mtile = make_tile_mask(tile_segbits[tile_type], tile_name, tile_bits)
        verbose and print(
            "Checking %s, type %s, bits: %s" %
            (tile_name, tile_type, len(mtile)))
        if len(mtile) == 0:
            continue

        collisions = set()
        for bits in mtile.keys():
            if bits in mall.keys():
                collisions.add(bits)

        if collisions:
            print("ERROR: %s collisions" % len(collisions))
            for ck in sorted(collisions):
                addr, bitaddr = ck
                word, bit = util.addr_bit2word(bitaddr)
                print(
                    "  %s: had %s, got %s" %
                    (util.addr2str(addr, word, bit), mall[ck], mtile[ck]))
            raise ValueError("%s collisions" % len(collisions))
        mall.update(mtile)
        tiles_checked += 1
    print("Checked %s tiles, %s bits" % (tiles_checked, len(mall)))


def run(db_root, part, verbose=False):
    # Start by running a basic check on db files
    print("Checking individual .db...")
    parsedb_all(db_root, verbose=verbose)

    # Now load and verify tile consistency
    db = prjxraydb.Database(db_root, part)
    db._read_tilegrid()
    '''
    these don't load properly without .json files
    See: https://github.com/SymbiFlow/prjxray/issues/303
    db._read_tile_types()
    print(db.tile_types.keys())
    '''

    verbose and print("")

    print("Checking aggregate dir...")
    check_tile_overlap(db, verbose=verbose)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Parse a db repository, checking for consistency")

    util.db_root_arg(parser)
    util.part_arg(parser)
    parser.add_argument('--verbose', action='store_true', help='')
    args = parser.parse_args()

    run(args.db_root, args.part, verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker
from prjxray import verilog


def bitfilter(frame, word):
    if frame < 28:
        return False

    return True


def bus_tags(segmk, ps, all_params, site):
    segmk.add_site_tag(site, 'IN_USE', ps['active'])

    if not ps['active']:
        return

    params = all_params[site]["params"]

    #for k in ps:
    #    segmk.add_site_tag(site, 'param_' + k + '_' + str(ps[k]), 1)

    for reg, invert in [
        ('RST', 1),
        ('PWRDWN', 1),
        ('CLKINSEL', 0),
        ('PSEN', 1),
        ('PSINCDEC', 1),
    ]:
        opt = 'IS_{}_INVERTED'.format(reg)

        if invert:
            segmk.add_site_tag(site, 'ZINV_' + reg, 1 ^ ps[opt])
        else:
            segmk.add_site_tag(site, 'INV_' + reg, ps[opt])

    for opt in ['OPTIMIZED', 'HIGH', 'LOW']:
        if verilog.unquote(ps['BANDWIDTH']) == opt:
            segmk.add_site_tag(site, 'BANDWIDTH.' + opt, 1)
        elif verilog.unquote(ps['BANDWIDTH']) == 'LOW':
            segmk.add_site_tag(site, 'BANDWIDTH.' + opt, 0)

    # "INTERNAL" compensation conflicts with the CLKFBOUT2IN->CLKFBIN PIP.
    # There is no telling which of these two is actually controlled by those
    # bits. It is better to leave them for the PIP.
    COMPENSATION_OPTS = ['ZHOLD', 'BUF_IN', 'EXTERNAL']

    for opt in COMPENSATION_OPTS:
        val = params["COMPENSATION"] == opt
        segmk.add_site_tag(site, "COMP.{}".format(opt), val)
        segmk.add_site_tag(site, "COMP.Z_{}".format(opt), not val)

    opt = (verilog.unquote(ps["SS_EN"]) == "TRUE")
    segmk.add_site_tag(site, "SS_EN", opt)

    for param in ['CLKFBOUT_MULT_F']:
        paramadj = int(ps[param])
        bitstr = [int(x) for x in "{0:09b}".format(paramadj)[::-1]]
        for i in range(7):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])

    for param in ['CLKOUT0_DUTY_CYCLE']:
        assert ps[param][:2] == '0.', ps[param]
        assert len(ps[param]) == 5
        paramadj = int(ps[param][2:])
        bitstr = [int(x) for x in "{0:011b}".format(paramadj)[::-1]]

        for i in range(10):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])

    for param, bits in [
        ('CLKOUT0_DIVIDE_F', 7),
        ('CLKOUT1_DIVIDE', 7),
        ('CLKOUT2_DIVIDE', 7),
        ('CLKOUT3_DIVIDE', 7),
        ('CLKOUT4_DIVIDE', 7),
        ('CLKOUT5_DIVIDE', 7),
        ('CLKOUT6_DIVIDE', 7),
        ('DIVCLK_DIVIDE', 6),
    ]:
        # 1-128 => 0-127 for actual 7 bit value
        paramadj = int(ps[param])
        if paramadj < 4:
            continue

        bitstr = [int(x) for x in "{0:09b}".format(paramadj)[::-1]]
        for i in range(bits):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])

    segmk.add_site_tag(
        site, 'STARTUP_WAIT',
        verilog.unquote(ps['STARTUP_WAIT']) == 'TRUE')


def run():

    segmk = Segmaker("design.bits")

    print("Loading params")
    f = open("params.json")
    params = json.load(f)
    params = {p["site"]: p for p in params}

    print("Loading tags")
    f = open('params.jl', 'r')
    f.readline()
    for l in f:
        j = json.loads(l)
        bus_tags(segmk, j, params, j['site'])

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import pyjson5
import simplejson
import sys


def main():
    simplejson.dump(pyjson5.load(sys.stdin), sys.stdout, indent=2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse


def main():
    parser = argparse.ArgumentParser(
        description="Converts generic BRAM pips to BRAM_L and BRAM_R pips.")

    parser.add_argument('--bram_x', required=True)
    parser.add_argument('--bram_l', required=True)
    parser.add_argument('--bram_r', required=True)

    args = parser.parse_args()

    with open(args.bram_x, 'r') as f_in, open(
            args.bram_l, 'w') as f_l_out, open(args.bram_r, 'w') as f_r_out:
        for l in f_in:
            # BRAM_L has the same pip names as BRAM_X
            print(l.strip(), file=f_l_out)

            # BRAM_R has some _R_ added to some pips.
            #
            # BRAM.BRAM_ADDRARDADDRL0.BRAM_IMUX_ADDRARDADDRL0
            #
            # becomes
            #
            # BRAM.BRAM_ADDRARDADDRL0.BRAM_IMUX_R_ADDRARDADDRL0
            print(
                l.strip().replace('BRAM_IMUX_ADDR', 'BRAM_R_IMUX_ADDR'),
                file=f_r_out)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import os
from enum import Enum

from prjxray.segmaker import Segmaker, add_site_group_zero

INT = "INT"
BIN = "BIN"
BOOL = "BOOL"
STR = "STR"


def bitfilter_gtx_channel_x(frame, bit):
    # Filter out interconnect bits.
    if frame not in [28, 29, 30, 31]:
       return False

    return True


def bitfilter_gtx_channel_x_mid(frame, bit):
    # Filter out interconnect bits.
    if frame not in [0, 1, 2, 3]:
        return False

    return True


def main():
    segmk = Segmaker("design.bits")

    fuz_dir = os.getenv("FUZDIR", None)
    assert fuz_dir
    with open(os.path.join(fuz_dir, "attrs.json"), "r") as attr_file:
        attrs = json.load(attr_file)

    print("Loading tags")
    with open("params.json") as f:
        primitives_list = json.load(f)

    for primitive in primitives_list:
        tile_type = primitive["tile_type"]
        params_list = primitive["params"]

        for params in params_list:
            site = params["site"]

            if "GTXE2_CHANNEL" not in site:
                continue

            in_use = params["IN_USE"]

            segmk.add_site_tag(site, "IN_USE", in_use)

            if in_use:
                for param, param_info in attrs.items():
                    value = params[param]
                    param_type = param_info["type"]
                    param_digits = param_info["digits"]
                    param_values = param_info["values"]

                    if param_type == INT:
                        param_encodings = param_info["encoding"]
                        param_encoding = param_encodings[param_values.index(
                            value)]
                        bitstr = [
                            int(x) for x in "{value:0{digits}b}".format(
                                value=param_encoding, digits=param_digits)
                            [::-1]
                        ]

                        for i in range(param_digits):
                            segmk.add_site_tag(
                                site, '%s[%u]' % (param, i), bitstr[i])
                    elif param_type == BIN:
                        bitstr = [
                            int(x) for x in "{value:0{digits}b}".format(
                                value=value, digits=param_digits)[::-1]
                        ]

                        for i in range(param_digits):
                            segmk.add_site_tag(
                                site, "%s[%u]" % (param, i), bitstr[i])
                    elif param_type == BOOL:
                        segmk.add_site_tag(site, param, value == "TRUE")
                    else:
                        assert param_type == STR

                        # The RXSLIDE_MODE parameter has overlapping bits
                        # for its possible values. We need to treat it
                        # differently
                        if param == "RXSLIDE_MODE":
                            add_site_group_zero(
                                segmk, site, "{}.".format(param), param_values,
                                "OFF", value)
                        else:
                            for param_value in param_values:
                                segmk.add_site_tag(
                                    site, "{}.{}".format(param, param_value),
                                    value == param_value)

                for param in ["TXUSRCLK", "TXUSRCLK2", "TXPHDLYTSTCLK",
                              "RXUSRCLK", "RXUSRCLK2", "DRPCLK"]:
                    segmk.add_site_tag(site, "INV_" + param, params[param])

    gtx_channel_x = [
        "GTX_CHANNEL_0",
        "GTX_CHANNEL_1",
        "GTX_CHANNEL_2",
        "GTX_CHANNEL_3",
    ]

    gtx_channel_x_mid = [
        "GTX_CHANNEL_0_MID_LEFT",
        "GTX_CHANNEL_1_MID_LEFT",
        "GTX_CHANNEL_2_MID_LEFT",
        "GTX_CHANNEL_3_MID_LEFT",
        "GTX_CHANNEL_0_MID_RIGHT",
        "GTX_CHANNEL_1_MID_RIGHT",
        "GTX_CHANNEL_2_MID_RIGHT",
        "GTX_CHANNEL_3_MID_RIGHT",
    ]

    if tile_type in gtx_channel_x:
        bitfilter = bitfilter_gtx_channel_x
    elif tile_type in gtx_channel_x_mid:
        bitfilter = bitfilter_gtx_channel_x_mid
    else:
        assert False, tile_type

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
class NodeModel():
    """ Node lookup model

    Terminology:
     Wire - A segment of metal in a tile
     Node - A connected set of wires

    This class can provide a list of nodes, the wires in a node and the node
    that a wire belongs too.

    The name of node is always the name of one wire in the node.

    It is recommended that this class be constructed by calling
    Database.node_model rather than constructing this class directly.

    """

    def __init__(
            self, grid, connections, tile_wires, node_wires, progressbar=None):
        self.grid = grid
        self.connections = connections
        self.tile_wires = tile_wires
        self.specific_node_wires = set(node_wires['specific_node_wires'])

        node_pattern_wires = node_wires['node_pattern_wires']
        self.node_pattern_wires = {}
        for tile_type in node_pattern_wires:
            assert tile_type not in self.node_pattern_wires
            self.node_pattern_wires[tile_type] = set(
                node_pattern_wires[tile_type])

        for tile_type in self.tile_wires:
            if tile_type not in self.node_pattern_wires:
                self.node_pattern_wires[tile_type] = set()

        self.nodes = None

        self.wire_to_node_map = None

        if progressbar is None:
            self.progressbar = lambda x: x
        else:
            self.progressbar = progressbar

    def _build_nodes(self):
        tile_wire_map = {}
        wires = {}
        flat_wires = []

        for tile in self.progressbar(self.grid.tiles()):
            gridinfo = self.grid.gridinfo_at_tilename(tile)
            tile_type = gridinfo.tile_type

            for wire in self.tile_wires[tile_type]:
                wire_pkey = len(flat_wires)
                tile_wire_map[(tile, wire)] = wire_pkey
                flat_wires.append((tile, wire))
                wires[wire_pkey] = None

        for connection in self.progressbar(self.connections.get_connections()):
            a_pkey = tile_wire_map[(
                connection.wire_a.tile, connection.wire_a.wire)]
            b_pkey = tile_wire_map[(
                connection.wire_b.tile, connection.wire_b.wire)]

            a_node = wires[a_pkey]
            b_node = wires[b_pkey]

            if a_node is None:
                a_node = set((a_pkey, ))

            if b_node is None:
                b_node = set((b_pkey, ))

            if a_node is not b_node:
                a_node |= b_node

                for wire in a_node:
                    wires[wire] = a_node

        nodes = {}
        for wire_pkey, node in self.progressbar(wires.items()):
            if node is None:
                node = set((wire_pkey, ))

            assert wire_pkey in node

            nodes[id(node)] = node

        def get_node_wire_for_wires(wire_pkeys):
            if len(wire_pkeys) == 1:
                for wire_pkey in wire_pkeys:
                    return flat_wires[wire_pkey]

            for wire_pkey in wire_pkeys:
                tile, wire = flat_wires[wire_pkey]

                if '{}/{}'.format(tile, wire) in self.specific_node_wires:
                    return tile, wire

            for wire_pkey in wire_pkeys:
                tile, wire = flat_wires[wire_pkey]

                gridinfo = self.grid.gridinfo_at_tilename(tile)

                if wire in self.node_pattern_wires[gridinfo.tile_type]:
                    return tile, wire

            return None

        self.nodes = {}
        for node_wire_pkeys in self.progressbar(nodes.values()):
            node_wire = get_node_wire_for_wires(node_wire_pkeys)
            if node_wire is None:
                continue

            self.nodes[node_wire] = [
                flat_wires[wire_pkey] for wire_pkey in node_wire_pkeys
            ]

    def get_nodes(self):
        """ Return a set of node names. """
        if self.nodes is None:
            self._build_nodes()

        return self.nodes.keys()

    def get_wires_for_node(self, tile, wire):
        """ Get wires in node named for specified tile and wire. """
        if self.nodes is None:
            self._build_nodes()

        return self.nodes[tile, wire]

    def _build_wire_to_node_map(self):
        self.wire_to_node_map = {}

        if self.nodes is None:
            self._build_nodes()

        for node, wires in self.nodes.items():
            for tile_wire in wires:
                assert tile_wire not in self.wire_to_node_map
                self.wire_to_node_map[tile_wire] = node

    def get_node_for_wire(self, tile, wire):
        """ Get node for specified tile and wire. """
        if self.wire_to_node_map is None:
            self._build_wire_to_node_map()

        return self.wire_to_node_map[tile, wire]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

tiledata = dict()
pipdata = set()

print("Loading tags from design.txt.")
with open("design.txt", "r") as f:
    for line in f:
        ab, dst, src = line.split()
        tile, dst = dst.split("/")
        _, src = src.split("/")

        if tile not in tiledata:
            tiledata[tile] = {
                "pips": set(),
                "nodes": set(),
            }

        if ab == "A":
            tiledata[tile]["pips"].add((dst, src))
            pipdata.add((dst, src))
        else:
            tiledata[tile]["nodes"].add(src)
            tiledata[tile]["nodes"].add(dst)

for tile, pips_nodes in tiledata.items():
    pips = pips_nodes["pips"]
    nodes = pips_nodes["nodes"]

    for dst, src in pipdata:
        if (dst, src) in pips:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 1)
        elif dst not in nodes and src not in nodes:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 0)


def bitfilter(frame_idx, bit_idx):
    return frame_idx in [0, 1]


segmk.compile(bitfilter=bitfilter)
segmk.write(allow_empty=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Route timing delay definitions.
Routing delay is formed from two parts in this model:

- Intristic delay of the element
- Capactive loading delay of the net

Intristic delay is a time value (e.g. nanoseconds), does not vary based on
routing fanout.  It does vary based on the PVT (process, voltage, temperature)
corner.  PvtCorner and IntristicDelay objects are used to model intristic
delay of elements.

Capactive loading is the elmore delay from the RC tree formed by interconnect.
The RC tree is made up of 5 types of RC nodes:


|Element type          |Object        |Intrinsic delays?|Output resistance?|Capacitance type    |
|----------------------|--------------|-----------------|------------------|--------------------|
|Site output pin       |Outpin        |Yes              |Yes               |N/A                 |
|Buffered switch       |Buffer        |Yes              |Yes               |Internal capacitance|
|Pass-transistor switch|PassTransistor|Yes              |Yes               |N/A                 |
|Wire                  |Wire          |No               |Yes               |Pi model            |
|Site input pin        |Inpin         |Yes              |No                |Input capacitance   |

The elmore delay is the RC tree formed by these 5 components.  Out pins and
buffer switches are the roots of the elmore tree. Buffer switches
and inpins are leafs of the elmore tree. Wires and pass-transistor switches are
nodes in the tree. Wires share their capacitance upstream and downstream using
a pi-model.

Example timing tree:

    +------+
    |Outpin|
    +--+---+
    |
    |
    v
    +--+--+
    |Wire |
    +--+--+
    |
    +-----------------+
    |                 |
    +--+---+     +-------+------+
    |Buffer|     |PassTransistor|
    +--+---+     +------+-------+
    |                |
    v                v
    +--+-+           +--+-+
    |Wire|           |Wire|
    +--+-+           +--+-+
    |                |
    v                v
    +--+--+          +--+--+
    |Inpin|          |Inpin|
    +-----+          +-----+

Note on units:

The timing model operates on the following types of units:
 - Time
 - Resistance
 - Capacitance

For a consistent unit set, the following equation must be satisfied:

1 Resistance unit * 1 Capacitance unit = 1 Time unit

The SI unit set would be:
 - Time = seconds
 - Resistance = Ohms
 - Capacitance = Farads

However as long as the scale factors are consistent, the model will work
with other unit combinations.  For example:
 - Time = nanoseconds (1e-9 seconds)
 - Resistance = milliOhms (1e-3 Ohms)
 - Capacitance = microFarads (1e-6 Farads)

(1e-3 * 1e-6) (Ohms * Farads) does equal (1e-9) seconds.

"""
import enum
from collections import namedtuple


class PvtCorner(enum.Enum):
    """ Process/voltage/temperature corner definitions. """

    # Corner where device operates with fastest intristic delays.
    FAST = "FAST"

    # Corner where device operates with slowest intristic delays.
    SLOW = "SLOW"

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class IntristicDelay(namedtuple('IntristicDelay', 'min max')):
    """ An intristic delay instance.

    Represents is the intristic delay through an element (e.g. a site pin or
    interconnect pip).

    The intristic delay of an element is generally modelled at a particular
    "corner" of a design.  The "corner" generally speaking is modelled over
    process, voltage and temperature PVT.  The IntristicDelay object
    reperesents the minimum or maximum delay through all instances of the
    element at 1 corner.

    Attributes
    ----------

    min : float
        Minimum instrinsic delay (nsec)

    max : float
        Maximum instrinsic delay (nsec)
    """


class RcElement(namedtuple('RcElement', 'resistance capacitance')):
    """ One part of an RcNode, embedded within an RcTree.

    Attributes
    ----------

    resistance : float
        Resistance of element

    capacitance : float
        Capacitance of element

    """
    pass


class hashabledict(dict):
    """ Immutable version of dictionary with hash support. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hash = hash(tuple(sorted(self.items())))

    def setdefault(self, *args, **kwargs):
        raise NotImplementedError("hashabledict cannot be mutated.")

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError("hashabledict cannot be mutated.")

    def update(self, *args, **kwargs):
        raise NotImplementedError("hashabledict cannot be mutated.")

    def __hash__(self):
        return self.hash


def fast_slow_tuple_to_corners(arr):
    """ Convert delay 4-tuple into two IntristicDelay objects.

    Returns
    -------

    corners : dict of PvtCorner to IntristicDelay
        Dictionary keys of FAST and SLOW, mapping to the instrinsic delay
        for each corner.

    """

    fast_min, fast_max, slow_min, slow_max = map(float, arr)

    return hashabledict(
        {
            PvtCorner.FAST: IntristicDelay(
                min=fast_min,
                max=fast_max,
            ),
            PvtCorner.SLOW: IntristicDelay(
                min=slow_min,
                max=slow_max,
            ),
        })


class TimingNode(object):
    """ Base class for timing node models.
    """

    def get_intrinsic_delays(self):
        """ Returns Intristic delays (if any) timing node.

        Returns
        -------

        Dictionary of PvtCorner to Intristic. Is None if node has no intristic
        delay.
        """

        pass

    def get_rc_delay(self):
        """ Return portion of net delay due to elmore (RC) delay at this node.

        Must be called after propigate_delays has been called on the Outpin
        object of this tree.

        """
        pass

    def get_downstream_cap(self):
        """ Returns downstream capacitance at this node.

        Must be called after propigate_delays has been called on the Outpin
        object of this tree.

        """
        pass

    def propigate_downstream_capacitance(self, math):
        """ Returns capacitance visible to parent of this node.

        Must call propigate_downstream_capacitance on all children of this node.
        Should save downstream capacitance visible to this node's output
        (if any) to be returned in the get_downstream_cap method.
        """
        pass


class DownstreamNode(TimingNode):
    """ All non-root TimingNode's are DownstreamNode's.

    """

    def propigate_delays(self, elements, math):
        """ Propigates upstream delay elements to children of the tree.

        Must call propigated_delays on all children of this node, and add this
        node to elements.

        Arguments
        ---------
        elements : list of TimingNode's
            List of delay nodes between root of this tree and this node.
        math : MathModel
            Math model to use to compute delays
        """
        pass


class Outpin(TimingNode):
    """ Represents a site output pin.

    Outpin object is the root of the timing tree.  Once tree is built with
    set_sink_wire and Wire.add_child methods, propigate_delays should be
    invoked to estabilish model.

    Arguments
    ---------

    resistance
        Drive resistance in elmore delay model
    delays
        Intristic delays on output pin.

    """

    def __init__(self, resistance, delays):
        self.resistance = resistance
        self.delays = delays
        self.sink_wire = None

        self.downstream_cap = None
        self.rc_delay = None

    def set_sink_wire(self, wire):
        """ Sets sink wire for this output pin.

        An output pin always sinks to exactly 1 wire.

        This method must be called prior to calling propigate_delays method
        on this object.

        Arguments
        ---------
        wire : Wire object
            Sink wire for this output pin.

        """
        self.sink_wire = wire

    def propigate_downstream_capacitance(self, math):
        assert self.sink_wire is not None
        self.downstream_cap = self.sink_wire.propigate_downstream_capacitance(
            math)
        self.rc_delay = math.multiply(self.downstream_cap, self.resistance)

    def propigate_delays(self, math):
        """ Propigate delays throughout tree using specified math model.

        Must be called after elmore tree is estabilished.

        Arguments
        ---------
        math : MathModel object
            Math model used when doing timing computations.

        """
        self.propigate_downstream_capacitance(math)
        self.sink_wire.propigate_delays([self], math)
        self.rc_delay = math.multiply(self.resistance, self.downstream_cap)

    def get_intrinsic_delays(self):
        return self.delays

    def get_rc_delay(self):
        assert self.rc_delay is not None
        return self.rc_delay

    def get_downstream_cap(self):
        assert self.downstream_cap is not None
        return self.downstream_cap


class Inpin(DownstreamNode):
    """ Represents a site input pin.

    Represents leaf of timing model.  Once model is connected and delays
    are propigate (by calling Outpin,propigated_delays), get_delays will
    correctly return the list of delay elements from the root to this leaf.

    Arguments
    ---------

    capacitance
        Pin capacitance for input pin.
    delays
        Intristic delays on input pin.

    """

    def __init__(self, capacitance, delays, name=None):
        self.capacitance = capacitance
        self.delays = delays
        self.propigated_delays = None
        self.name = name

    def get_intrinsic_delays(self):
        return self.delays

    def get_rc_delay(self):
        return None

    def get_downstream_cap(self):
        return None

    def propigate_downstream_capacitance(self, math):
        return self.capacitance

    def propigate_delays(self, elements, math):
        self.propigated_delays = list(elements)

    def get_delays(self):
        """ Return list of delay models that make up the delay for this pin.

        The sum of all delay elements (both intristic and RC) is the net
        delay from the output pin to this input pin.

        """
        return self.propigated_delays + [self]


class Wire(DownstreamNode):
    """ Represents a wire in the timing model.

    Wires must be connected to an upstream driver model (Outpin, Buffer,
    PassTransistor objects) with set_sink_wire, and add_child must be called
    attaching output nodes (Buffer, PassTransistor, Inpin objects).

    Arguments
    ---------
    rc_elements : List of RcElement
        Resistance and capacitance of this wire.
    math : MathModel
        Math model used to compute lumped resistance and capacitance.

    """

    def __init__(self, rc_elements, math):
        self.resistance = math.sum(elem.resistance for elem in rc_elements)
        self.capacitance = math.sum(elem.capacitance for elem in rc_elements)
        self.children = []

        self.downstream_cap = None
        self.propigated_delays = None
        self.rc_delay = None

    def add_child(self, child):
        """ Add a child node to this wire.

        Call this method as needed prior to calling propigate_delays on the
        root Outpin object.

        Arguments
        ---------
        child : Buffer or PassTransistor or Inpin
            Adds child load to this wire.

        """
        self.children.append(child)

    def propigate_downstream_capacitance(self, math):
        downstream_cap = math.sum(
            child.propigate_downstream_capacitance(math)
            for child in self.children)

        # Pi-model is definied such that wire resistance only sees half of the
        # wire capacitance.
        self.downstream_cap = math.plus(
            math.divide(self.capacitance, 2), downstream_cap)

        # Upstream seems all of the wires capacitance
        return math.plus(downstream_cap, self.capacitance)

    def propigate_delays(self, elements, math):
        self.propigated_delays = list(elements)

        for child in self.children:
            child.propigate_delays(self.propigated_delays + [self], math)

        self.rc_delay = math.multiply(self.resistance, self.downstream_cap)

    def get_intrinsic_delays(self):
        return None

    def get_rc_delay(self):
        assert self.rc_delay is not None
        return self.rc_delay

    def get_downstream_cap(self):
        assert self.downstream_cap is not None
        return self.downstream_cap


class Buffer(DownstreamNode):
    """ Represents an isolating switch.

    The internal_capacitance model is such that the upstream node only sees
    the capacitance of this node when the switch is enabled.  Therefore, only
    active buffers should be included in the model.

    Arguments
    ---------
    internal_capacitance
        Capacitance seen by upstream node when this buffer is enabled.
    drive_resistance
        Driver resistance used for computing elmore delay.
    delays : Dictionary of PvtCorner to IntristicDelay
        Delay through switch

    """

    def __init__(self, internal_capacitance, drive_resistance, delays):
        self.internal_capacitance = internal_capacitance
        self.drive_resistance = drive_resistance
        self.delays = delays

        self.downstream_cap = None
        self.rc_delay = None

    def set_sink_wire(self, wire):
        """ Sets sink wire for this output pin.

        An output pin always sinks to exactly 1 wire.

        This method must be called prior to calling propigate_delays method
        on the root Outpin object of this tree.

        Arguments
        ---------
        wire : Wire object
            Sink wire for this output pin.

        """
        self.sink_wire = wire

    def propigate_downstream_capacitance(self, math):
        assert self.sink_wire is not None
        self.downstream_cap = self.sink_wire.propigate_downstream_capacitance(
            math)
        return self.internal_capacitance

    def propigate_delays(self, elements, math):
        self.propigated_delays = list(elements)

        assert self.sink_wire is not None
        self.sink_wire.propigate_delays(self.propigated_delays + [self], math)
        self.rc_delay = math.multiply(
            self.downstream_cap, self.drive_resistance)

    def get_intrinsic_delays(self):
        return self.delays

    def get_rc_delay(self):
        assert self.rc_delay is not None
        return self.rc_delay

    def get_downstream_cap(self):
        assert self.downstream_cap is not None
        return self.downstream_cap


class PassTransistor(DownstreamNode):
    """ Represents a non-isolating switch.

    Arguments
    ---------
    drive_resistance
        Driver resistance used for computing elmore delay.
    delays : Dictionary of PvtCorner to IntristicDelay
        Delay through switch.

    """

    def __init__(self, drive_resistance, delays):
        self.drive_resistance = drive_resistance
        self.delays = delays
        self.sink_wire = None

        self.downstream_cap = None
        self.rc_delay = None

    def set_sink_wire(self, wire):
        """ Sets sink wire for this output pin.

        An output pin always sinks to exactly 1 wire.

        This method must be called prior to calling propigate_delays method
        on the root Outpin object of this tree.

        Arguments
        ---------
        wire : Wire object
            Sink wire for this output pin.

        """
        self.sink_wire = wire

    def propigate_downstream_capacitance(self, math):
        assert self.sink_wire is not None
        self.downstream_cap = self.sink_wire.propigate_downstream_capacitance(
            math)

        return self.downstream_cap

    def propigate_delays(self, elements, math):
        self.propigated_delays = list(elements)

        assert self.sink_wire is not None
        self.sink_wire.propigate_delays(self.propigated_delays + [self], math)
        self.rc_delay = math.multiply(
            self.downstream_cap, self.drive_resistance)

    def get_intrinsic_delays(self):
        return self.delays

    def get_rc_delay(self):
        assert self.rc_delay is not None
        return self.rc_delay

    def get_downstream_cap(self):
        assert self.downstream_cap is not None
        return self.downstream_cap



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys
import time
import argparse

from litex import RemoteClient

from sdram_init import *


def identify_fpga(wb):
    """ Gets the FPGA identifier and prints it on terminal."""
    fpga_id = ""
    for i in range(256):
        c = chr(wb.read(wb.bases.identifier_mem + 4 * i) & 0xff)
        fpga_id += c
        if c == "\0":
            break
    print(fpga_id)


def seed_to_data(seed, random=True):
    if random:
        return (1664525 * seed + 1013904223) & 0xffffffff
    else:
        return seed


def write_pattern(wb, length):
    for i in range(length):
        wb.write(wb.mems.main_ram.base + 4 * i, seed_to_data(i))


def check_pattern(wb, length, debug=False):
    errors = 0
    for i in range(length):
        error = 0
        if wb.read(wb.mems.main_ram.base + 4 * i) != seed_to_data(i):
            error = 1
            if debug:
                print(
                    "{}: 0x{:08x}, 0x{:08x} KO".format(
                        i, wb.read(wb.mems.main_ram.base + 4 * i),
                        seed_to_data(i)))
        else:
            if debug:
                print(
                    "{}: 0x{:08x}, 0x{:08x} OK".format(
                        i, wb.read(wb.mems.main_ram.base + 4 * i),
                        seed_to_data(i)))
        errors += error
    return errors


def find_bitslips_delays(wb):
    """ Finds bitslip and delay values that can be used with the implemented DDR design."""
    nbitslips = 8
    ndelays = 32
    nmodules = 2
    nwords = 16

    final_bitslip = None
    final_delay = None

    for bitslip in range(nbitslips):
        print("bitslip {:d}: |".format(bitslip), end="")
        for delay in range(ndelays):
            for module in range(nmodules):
                wb.regs.ddrphy_dly_sel.write(1 << module)
                wb.regs.ddrphy_rdly_dq_rst.write(1)
                wb.regs.ddrphy_rdly_dq_bitslip_rst.write(1)
                for i in range(bitslip):
                    wb.regs.ddrphy_rdly_dq_bitslip.write(1)
                for i in range(delay):
                    wb.regs.ddrphy_rdly_dq_inc.write(1)
            write_pattern(wb, nwords)
            errors = check_pattern(wb, nwords)
            if errors:
                print("..|", end="")
            else:
                print("{:02d}|".format(delay), end="")
                final_bitslip = bitslip if final_bitslip is None else final_bitslip
                final_delay = delay if final_delay is None else final_delay
            sys.stdout.flush()
        print("")

    assert final_bitslip is not None and final_delay is not None, "bitslip/delay values not found"

    return final_bitslip, final_delay


def set_bitslip_delay(wb, bitslip, delay):
    """ Sets bitslip and delay values."""
    nmodules = 2

    for module in range(nmodules):
        wb.regs.ddrphy_dly_sel.write(1 << module)
        wb.regs.ddrphy_rdly_dq_rst.write(1)
        wb.regs.ddrphy_rdly_dq_bitslip_rst.write(1)
        for i in range(bitslip):
            wb.regs.ddrphy_rdly_dq_bitslip.write(1)
        for i in range(delay):
            wb.regs.ddrphy_rdly_dq_inc.write(1)


def read_word_offset(read_only=False):
    word = None
    if not read_only:
        print("\n==================================================\n")
        print(
            "Set a byte long word to write to memory (e.g. 0xdeadbeef): ",
            end="")
        word = int(input(), 16) & 0xffffffff

    print("\n==================================================\n")
    print("Set offset from base memory address: ", end="")
    offset = int(input())

    print("\n==================================================\n")
    print("Set number of words to read or write: ", end="")
    length = int(input())

    return word, offset, length


def write_user_pattern(wb, offset, length, pattern):
    for i in range(length):
        wb.write(wb.mems.main_ram.base + 4 * (offset + i), pattern)


def read_user_pattern(wb, offset, length, pattern=None):
    for i in range(length):
        read_value = wb.read(wb.mems.main_ram.base + 4 * (offset + i))

        if pattern is None:
            print("0x{:08x}".format(read_value))
        else:
            if read_value == pattern:
                outcome = "CORRECT"
            else:
                outcome = "INCORRECT"

            print(
                "{} --> 0x{:08x}, 0x{:08x}".format(
                    outcome, read_value, pattern))


def start_command_interface(wb):

    cmd_list = """
Commands list:
    0 --> Write memory
    1 --> Read memory
    2 --> Write/Read memory
    3 --> Print commands list
    4 --> Exit
"""

    print(cmd_list)

    while True:
        print("\nWaiting for command: ", end="")
        cmd = int(input())

        if cmd == 0:
            word, offset, length = read_word_offset()
            write_user_pattern(wb, offset, length, word)
        elif cmd == 1:
            word, offset, length = read_word_offset(True)
            read_user_pattern(wb, offset, length)
        elif cmd == 2:
            word, offset, length = read_word_offset()
            write_user_pattern(wb, offset, length, word)
            read_user_pattern(wb, offset, length, word)
        elif cmd == 3:
            print(cmd_list)
        elif cmd == 4:
            break
        else:
            print("Command not recognized, try again...")


def main():
    parser = argparse.ArgumentParser(
        description="Script to test correct DDR behaviour.")

    parser.add_argument(
        '--bitslip', default=None, help="Defines a bitslip value.")
    parser.add_argument('--delay', default=None, help="Defines a delay value.")

    args = parser.parse_args()

    wb = RemoteClient(debug=False)
    wb.open()

    # software control
    wb.regs.sdram_dfii_control.write(0)

    # sdram initialization
    for i, (comment, a, ba, cmd, delay) in enumerate(init_sequence):
        print(comment)
        wb.regs.sdram_dfii_pi0_address.write(a)
        wb.regs.sdram_dfii_pi0_baddress.write(ba)
        if i < 2:
            wb.regs.sdram_dfii_control.write(cmd)
        else:
            wb.regs.sdram_dfii_pi0_command.write(cmd)
            wb.regs.sdram_dfii_pi0_command_issue.write(1)

    # hardware control
    wb.regs.sdram_dfii_control.write(dfii_control_sel)

    if args.bitslip is None or args.delay is None:
        bitslip, delay = find_bitslips_delays(wb)
    else:
        bitslip = int(args.bitslip)
        delay = int(args.delay)

    set_bitslip_delay(wb, bitslip, delay)

    start_command_interface(wb)

    wb.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.db import Database

BIN = "BIN"
BOOL = "BOOL"


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        if gridinfo.tile_type not in ["PCIE_BOT"]:
            continue

        for site_name, site_type in gridinfo.sites.items():

            return site_name, site_type


def main():
    print('''
module top();
''')

    lines = []

    site_name, site_type = gen_sites()

    params = dict()
    params['site'] = site_name

    verilog_attr = ""

    verilog_attr = "#("

    fuz_dir = os.getenv("FUZDIR", None)
    assert fuz_dir
    with open(os.path.join(fuz_dir, "attrs.json"), "r") as attrs_file:
        attrs = json.load(attrs_file)

    for param, param_info in attrs.items():
        param_type = param_info["type"]
        param_values = param_info["values"]
        param_digits = param_info["digits"]

        if param_type == BOOL:
            value = random.choice(param_values)
            value_str = verilog.quote(value)
        elif param_type == BIN:
            if type(param_values) is int:
                value = param_values
            elif type(param_values) is list:
                if len(param_values) > 1:
                    value = random.choice(param_values)
                else:
                    value = random.randint(0, param_values[0])
            value_str = "{digits}'b{value:0{digits}b}".format(
                value=value, digits=param_digits)
        params[param] = value

        verilog_attr += """
            .{}({}),""".format(param, value_str)

    verilog_attr = verilog_attr.rstrip(",")
    verilog_attr += "\n)"

    print("PCIE_2_1 {} pcie ();".format(verilog_attr))
    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import sys
import json
from prjxray.xjson import pprint
from prjxray.util import OpenSafeFile

if __name__ == "__main__":
    if len(sys.argv) == 1:
        import doctest
        doctest.testmod()
    else:
        assert len(sys.argv) == 2
        with OpenSafeFile(sys.argv[1]) as f:
            d = json.load(f)
        pprint(sys.stdout, d)



LVCMOS       = ['LVCMOS12', 'LVCMOS15', 'LVCMOS18']

# TODO: add support for digitally controlled impedance (_DCI) and termination (_T)
SSTL         = ['SSTL15',  #'SSTL15_DCI',  'SSTL15_T_DCI'
                'SSTL135', #'SSTL135_DCI', 'SSTL135_T_DCI'
                'SSTL12',  #'SSTL12_DCI',  'SSTL12_T_DCI'
               ]

# TODO: add support for digitally controlled impedance (_DCI) and termination (_T)
DIFF_SSTL15  = ['DIFF_SSTL15',  ] # 'DIFF_SSTL15_DCI',  'DIFF_SSTL15_T_DCI']
DIFF_SSTL135 = ['DIFF_SSTL135', ] # 'DIFF_SSTL135_DCI', 'DIFF_SSTL135_T_DCI']
DIFF_SSTL12  = ['DIFF_SSTL12',  ] # 'DIFF_SSTL12_DCI',  'DIFF_SSTL12_T_DCI']
DIFF_SSTL    = DIFF_SSTL15 + DIFF_SSTL135 + DIFF_SSTL12

LVDS         = ['LVDS']

DIFF         = DIFF_SSTL + LVDS